-- |Das Modul enthlt eine Funktion @solveBacktrack@ zur Lsung von Problemen
-- mittels backtracking.
--
-- Aus: Fethi Rabhi, Guy Lapalme, "Algorithms - a functional approach", p. 162ff.
module Gersdorf.Algorithms.Backtracking (
    solveBacktrack,
    solveBacktrackSolution,
    solveBacktrackStatistic,
    BacktrackingStatistic(..)
) where

import Gersdorf.Data.TreeFiniteMap
import Gersdorf.Data.ListStack
import Gersdorf.Data.BacktrackingChoice

import Data.Maybe
import Debug.Trace
import System.IO.Unsafe (unsafePerformIO)


solveBacktrack :: (Eq node, Show node) =>
    (node -> [node])    {- ^ Funktion, die ausgehend von einem Knoten n neue 
                             Knoten(-mglichkeiten) berechnet -}
    -> (node -> Bool)   -- ^ Funktion, die berechnet, ob ein Knoten eine Lsung ist
    -> node             -- ^ aktueller Knoten
    -> [node]           -- ^ Liste der Lsungsknoten

solveBacktrack succ goal cur = search (push cur (emptyStack::ListStack node))
    where
    search st 
        | isStackEmpty st = []
        | goal (top st)   = top st:(search (pop st))
        | otherwise       = let next = top st
                            in  search (foldr push (pop st) (succ next))

-------------------------------------------------------------------------------------------------

solveBacktrackSolution :: 
    (node -> BacktrackingChoice node)  {- ^ Funktion, die ausgehend von einem Knoten n neue 
                                            Knoten(-mglichkeiten) berechnet -}
    -> (node -> Bool)   -- ^ Funktion, die berechnet, ob ein Knoten eine Lsung ist
    -> (node -> g)      -- ^ Funktion, die einen Ergebnisknoten umrechnet.
    -> node             -- ^ aktueller Knoten
    -> [g]              -- ^ Liste der Lsungsknoten
solveBacktrackSolution getSucc isGoal transformGoal cur =
    map transformGoal
    $ filter isGoal 
    $ solveBacktrackR2 getSucc isGoal cur

-------------------------------------------------------------------------------------------------

solveBacktrackStatistic :: (Ord transformType) =>
    (node -> BacktrackingChoice node)  {- ^ Funktion, die ausgehend von einem Knoten n neue 
                                            Knoten(-mglichkeiten) berechnet -}
    -> (node -> Bool)   -- ^ Funktion, die berechnet, ob ein Knoten eine Lsung ist
    -> (node -> transformType) -- ^ Funktion, die aus Knoten den Typ der Transformation berechnet,
                               --   aus der der Knoten hervorgegangen ist
    -> node             -- ^ aktueller Knoten
    -> BacktrackingStatistic transformType   -- ^ Statistik
solveBacktrackStatistic getSucc isGoal getTransformType cur = 
    let nodeStream       = solveBacktrackR2 getSucc isGoal cur
        emptyBTStatistic = BacktrackingStatistic { btsTransformMap = emptyFM, btsSteps = 0, btsNoSolutions = 0 }

        calculateStatistic [] btStatistic = {-# SCC "1calcStatistic" #-} btStatistic
        calculateStatistic (node:nodes) btStatistic
            | isGoal node = {-# SCC "2calcStatistic" #-}
                let goalStatistic = nextStatistic { btsNoSolutions = 1 + (btsNoSolutions nextStatistic) }
                in  (calculateStatistic  nodes) $! goalStatistic
            | otherwise = {-# SCC "3calcStatistic" #-}
                (calculateStatistic nodes) $! nextStatistic
                
            where
            nextTransformType = getTransformType node
            nextStatistic = (updateBTStatistic nextTransformType) $! btStatistic
        
    in  calculateStatistic nodeStream emptyBTStatistic

updateBTStatistic :: (Ord transformType) => 
    transformType
    -> BacktrackingStatistic transformType
    -> BacktrackingStatistic transformType
updateBTStatistic tt bts =
    let newMap = (updateFM tt 1 (\i -> ((+) 1) $! i)) $! (btsTransformMap bts)
    in  BacktrackingStatistic { 
            btsSteps = (1 + btsSteps bts), 
            btsTransformMap = {-# SCC "btsTransformMap" #-} newMap,
            btsNoSolutions = btsNoSolutions bts
        }

data (Ord transformType) => BacktrackingStatistic transformType =
    BacktrackingStatistic {
        -- Die Striktheitsannotation fr btsTransformMap ist von essentieller
        -- Bedeutung fr den Speicherbedarf. Sie fhrt dazu, dass sich
        -- nicht "unendliche Berge" von unausgewerteten Thunks aufbauen, die
        -- bei der Ermittlung von Statistikinformationen berhaupt nicht
        -- bentigt werden.
        btsTransformMap :: !(TreeFiniteMap transformType Int),
        btsSteps        :: !Int,
        btsNoSolutions  :: !Int
    } deriving Show


--------------------------------------------------------------------------------------------
{- Erweiterung des Backtrackings. Wir betrachten nun Backtracking-Funktionen,
   deren Nachfolgesuchfunktion zwei Typen von Nachfolgeknotenmengen berechnen kann:
      1) BTOne - Eine Menge von Nachfolgeknoten. Falls einer dieser Knoten zu
                 einer Lsung fhrt, mssen die anderen Knoten nicht mehr betrachtet
                 werden, da diese zu keinen anderen Lsungsknoten fhren werden.
      2) BTMany - Eine Menge von Nachfolgeknoten. Falls einer der Knoten zu
                 einer Lsung fhrt, mssen die anderen dennoch betrachtet werden,
                 da diese zu anderen Lsungsknoten fhren knnen.

  Dementsprechend erhlt jeder Knoten auf dem Stack einen Marker, der besagt, ob
  eines seiner Kinder bereits zu einer Lsung gefhrt hat oder nicht. Initial
  ist dieser Marker auf False gesetzt.

  BTOne und BTMany werden gemeinsam mit dem Marker als Multiknoten bezeichnet.

  Algorithmus:
  0) Lege Startknoten s als BTMany([s], False)-Multiknoten auf den Stack

  1) Falls Stack leer: Ende.

  2) Nimm Multiknoten N vom Stack.
     WENN (N nichtleer) DANN
       WENN (N vom Typ BTOne && N hat Marker True) DANN
         propagateStackSolution
         goto 1)
       SONST
         Nimm Knoten n aus N
         WENN (n Zielknoten) DANN
           n zu Ergebnis zufgen
           Marker von N auf True setzen
           N auf Stack
           goto 1)
         SONST
           N auf Stack
           Nachfolgeknotenmenge von n berechnen
           Nachfolgeknotenmenge mit Marker False auf Stack legen
           goto 1)
     SONST {- N ist leer -}
       WENN (N hat Marker True) DANN
         propagateStackSolution
         goto 1)
       SONST
         goto 1)

  propagateStackSolution:
    WENN (Stack nichtleer) DANN
      nimm N2 vom Stack
      Setze Marker von N2 auf True
      lege N2 auf Stack

  Stack ist vom Typ :: ListStack (BacktrackingChoice node)
-}

solveBacktrackR2 :: 
    (node -> BacktrackingChoice node)  {- ^ Funktion, die ausgehend von einem Knoten n neue 
                                            Knoten(-mglichkeiten) berechnet -}
    -> (node -> Bool)   -- ^ Funktion, die berechnet, ob ein Knoten eine Lsung ist
    -> node             -- ^ aktueller Knoten
    -> [node]           -- ^ Liste der Lsungsknoten
solveBacktrackR2 getSucc isGoal cur = 
    search (push startNode (emptyStack::ListStack (BacktrackStackNode node))) 
    where
    startNode = BacktrackStackNode { bsnChoice = cBacktrackingDet cur, bsnHasSolution = False } 
    
    search st 
        | isStackEmpty st = {-trace "isStackEmpty"-} []
        | noChoice curBT  = 
            let nextStack = 
                    if (bsnHasSolution curStackChoice)
                        then propagateSolution (pop st)
                        else (pop st)
            in  {-trace ("NoChoice curBT")-} 
                search nextStack
        | (isBTOne curBT || isBTDet curBT) && (bsnHasSolution curStackChoice) =
            let nextStack = propagateSolution (pop st)
            in  {- trace (show nextTransformType ++ ": (isBTOne curBT || isBTDet curBT) && (bsnHasSolution curStackChoice)")-}
                search nextStack
        | isGoal curNode =
            let updStackNode = curStackChoice { bsnChoice = popChoice curBT, bsnHasSolution = True  }
            in  curNode:(search (push updStackNode st))
        | otherwise     = 
            let nextBT          = getSucc curNode
                updStackChoice  = curStackChoice { bsnChoice = popChoice curBT }
                nextStackChoice = BacktrackStackNode { bsnChoice = nextBT, bsnHasSolution = False }
            in  {-trace (show nextTransformType ++ ": otherwise")-}
                curNode:(search (push nextStackChoice $ push updStackChoice $ pop st))
        where
        curStackChoice = let c = top st
--                         in  trace (show c) c
                         in c
        curBT          = bsnChoice curStackChoice
        curNode        = fromJust $ topChoice curBT

        propagateSolution st = topUpdate (\bsn -> bsn { bsnHasSolution = True }) st

data BacktrackStackNode node = 
    BacktrackStackNode {
        bsnChoice :: BacktrackingChoice node,
        bsnHasSolution :: Bool
   } deriving Show 


