module Gersdorf.Algorithms.Random (
    pickFrom, pickFromIO, 
    pickFromWeighted, pickFromWeightedIO,
    distributeIO,
    distributeWeightedIO,
    module System.Random
) where 

import System.Random
import Data.Array.IO

import Gersdorf.Monad.IOUtil

-- picks one element from a given list by random
pickFrom :: (RandomGen g) => [a] -> g -> (a, g)
pickFrom l g = let (elnum, ng) = randomR (0, length l - 1) g
               in  ((l !! elnum), ng)

-- picks one element from a given list by random using the standard
-- random generator
pickFromIO :: [a] -> IO a
pickFromIO l = getStdRandom $ pickFrom l

pickFromWeighted :: (RandomGen g) => [(Int, a)] -> g -> (a, g)
pickFromWeighted l g = 
    let sumWeights = foldl (\c (w, _) -> c+w) 0 l
        firstVal   = case (head l) of
                         (_, v) -> v
        pickL      = tail $ scanl (\((_, c), _) (i, v) -> ((c, i+c), v)) ((0, 0), firstVal) l
        (selWeight, ng) = randomR (0, sumWeights-1) g
        
        findEl sw [] = error "end of picklist"
        findEl sw (((fw, tw), v):pls)
            | ((sw >= fw) && (sw < tw)) = v
            | otherwise                 = findEl sw pls
            
    in  (findEl selWeight pickL, ng)

pickFromWeightedIO :: [(Int, a)] -> IO a
pickFromWeightedIO l = getStdRandom $ pickFromWeighted l

-- | This function distributes r elements on n places by random.
distributeIO :: Int -> Int -> IO [Int]
distributeIO r n = 
     let weights = take n $ repeat 1
     in  distributeWeightedIO r weights

-- | Nimmt eine Anzahl von "Punkten", die auf eine Anzahl n von Stellen per
--   Zufall verteilt werden sollen. Die Anzahl der Stellen ergibt sich durch
--   die Lnge der bergebenen Liste. Die Liste enthlt Gewichtungen.
--   Beispiel: distributeWeighted 20 [1,2,3] verteilt 20 Punkte auf drei
--             Stellen. Dabei hat Stelle 1 die Gewichtung "1" und Stelle
--             3 die GEwichtung "3". Wird ein Punkt auf Stelle 3 verteilt,
--             so zhlt dieser Punkt dreifach.
--             Ein Beispielergebnis wre [1,3,4], da (1*1 + 2*3 + 3*4 = 19 < 20).
--   Es werden hchstens soviele Punkte vergeben, wie vorhanden sind, nie mehr!
distributeWeightedIO :: Int -> [Int] -> IO [Int]
distributeWeightedIO r ws = 
    let distribute' :: Int 
             -> Int                      -- minimum weight
             -> IOArray Int (Int, Int)   -- (weight, count)
             -> IO ()    
        distribute' r minWeight arr 
            | r < minWeight = 
                 return ()
            | otherwise = 
                 do let (_, n) = bounds arr
                    i <- randomRIO (1,n)
                    (w, v) <- readArray arr i
                    --putStr ("r=" ++ show r ++ ", i=" ++ show i ++ ", w=" ++ show w ++ "\n")
                    if (w > r) then distribute' r minWeight arr
                               else do writeArray arr i (w, v+1)
                                       distribute' (r-w) minWeight arr
        
        calcResult :: IOArray Int (Int, Int)
            -> IO [Int]
        calcResult arr =
            do el <- getElems arr
               return (map snd el)
                                                
    in  do let n =  length ws
           --putStr ("r=" ++ show r ++ "//w=" ++ show ws ++ "\n")
           arr   <- newListArray (1, n) (zip ws (repeat 1))  -- [(weight, count)]
           let alreadydistributed = foldl (+) 0 ws
               minimumWeight      = minimum ws
           
           distribute' (r-alreadydistributed) minimumWeight arr
           calcResult arr

