-- |This is the module containing a very simple implementation
-- of Terms.

module Gersdorf.ContextMatching.Base.ExtendedTerm (
    ExtendedTerm,
    ExtendedTermNode,
    StandardTreePath,
    toExtendedTerm,
    module Gersdorf.ContextMatching.Base.PositionTerm
) where

import Gersdorf.Data.ContextTree
import Gersdorf.ContextMatching.Base.Term 
import Gersdorf.ContextMatching.Base.PositionTerm
import Debug.Trace

-- | Ein ExtendedTerm ist ein ContextTree ber TermTypes
type ExtendedTerm = ExtendedTermT TermType
type ExtendedTermNode = ExtendedTermNodeT TermType

type ExtendedTermT tt = ContextTree ContextInfo tt
type ExtendedTermNodeT tt = TermNode (ExtendedTermT tt)

-- | Die Kontextinformationen.
data ContextInfo = ContextInfo
    { ciLeftHoles :: Int,  -- Anzahl Lcher links vom aktuellen Termknoten
      ciRightHoles :: Int, -- Anzahl Lcher rechts vom aktuellen Termknoten
      ciTotalHoles :: Int  -- Anzahl Lcher in diesem Subtermbazm
    } deriving Eq

instance (TermTypeable tt) => TermTypeable (ExtendedTermT tt) where
    -- Get Name of the root of this term
    getName t = getName (getType t)
    -- Holt Typ des aktuellen Wurzelknotens
    getType = getType . getContent

instance (TermTypeBuilder tt) => Term (ExtendedTermT tt) where
    -- gibt Anzahl der Lcher in diesem Subtermbaum zurck
    countHoles = getHoleCount

instance (TermTypeBuilder tt) => Linearizable (ExtendedTermT tt) where
    -- linearisiert den Term, d.h. alle Variablen und Kontextvariablen werden
    -- eindeutig umbenannt
    linearize ss t = 
        let transformStateF c t
                | isVariable t         = (c+1, TAThisNode (\_ -> cVar (newname t c)))
                | isContextVariable t  = (c+1, TAThisNode (\css -> cContextVar (newname t c) (head css)))
                | otherwise            = (c, TANothing)
            newname t c = (getName t) ++ (show c)
        in  transformTreeState transformStateF ss t

instance (TermTypeable tt) => TermIterator (ExtendedTermT tt)

instance (TermTypeBuilder tt) => Show (ExtendedTermT tt) where
    showsPrec _ t r = showTerm t r

instance (TermTypeBuilder tt) => SubTermBuilder (ExtendedTermT tt) where
    cVar n = cNodeC (cVarType n) cContextInfo [] 
    cFunction n tss = cNodeC (cFunctionType n) cContextInfo tss
    cContextVar n t = cNodeC (cContextVarType n) cContextInfo [t]
    cHole = cNodeC cHoleType cContextInfo []

cContextInfo :: (TermTypeBuilder tt) => ExtendedTermT tt -> ContextInfo
cContextInfo t = ContextInfo {
                    ciLeftHoles = calculateLeftHoles t,
                    ciRightHoles = calculateRightHoles t,
                    ciTotalHoles = calculateHoles t
                 }

calculateHoles :: (TermTypeBuilder tt) => ExtendedTermT tt -> Int
calculateHoles t | isHole t  = 1
                 | otherwise = foldl (\i c -> i+c) 0 (map getHoleCount (getChildren t))
--                 | otherwise = foldl (\a t -> (+) a $! getHoleCount t) 0 (getChildren t)

calculateLeftHoles :: ( TermTypeBuilder tt) => ExtendedTermT tt -> Int
calculateLeftHoles t = 
    case (getFather t) of
        Nothing  -> 0
        (Just f) -> process f (getChildIndex t)
    where
    process f ci | isContextVariable f = getLeftHoles f
                 | isFunction f        = let leftsiblings = fst (splitAt (ci-1) (getChildren f))
                                             leftholes    = foldl (\i c -> i+c) 0 (map getHoleCount leftsiblings)
                                         in  ((+) leftholes) $ (getLeftHoles f)
--                                             leftholes    = foldl (\i c -> seq ((+) i) c) 0 (map getHoleCount leftsiblings)
--                                         in  ((+) leftholes) $! (getLeftHoles f)

calculateRightHoles :: (TermTypeBuilder tt) => ExtendedTermT tt -> Int
calculateRightHoles t = 
    case (getFather t) of
        Nothing  -> 0
        (Just f) -> process f (getChildIndex t)
    where
    process f ci | isContextVariable f = getLeftHoles f
                 | isFunction f        = let rightsiblings = snd (splitAt ci (getChildren f))
                                             rightholes    = foldl (\i c -> i+c) 0 (map getHoleCount rightsiblings)
                                         in  ((+) rightholes) $ (getRightHoles f)
--                                             rightholes    = foldl (\i c -> seq ((+) i) c) 0 (map getHoleCount rightsiblings)
--                                         in  ((+) rightholes) $! (getRightHoles f)

instance ( TermTypeBuilder tt) => AnalysisTerm (ExtendedTermT tt) where
    getHoleCount = ciTotalHoles . getContext
    getLeftHoles = ciLeftHoles . getContext
    getRightHoles = ciRightHoles . getContext


instance ( TermTypeBuilder tt) => TermBuilder (ExtendedTermT tt) where
    -- Implementation of factory members
    cFunctionT n tss = cFunction n (map extractNode tss)
    cContextVarT n t = cContextVar n (extractNode t)

    cTerm = cTree

    replaceTerm ct ntn = replaceSubtree ct ntn

    extractSubTerm = cTree . extractNode


toExtendedTerm :: (Term t) => t -> ExtendedTerm
toExtendedTerm t = cTree $ toExtendedTerm' t
    where
    toExtendedTerm' t 
        | isHole t     = cHole
        | isVariable t = cVar (getName t)
        | isContextVariable t = cContextVar (getName t) (toExtendedTerm' $ head $ getChildren t)
        | isFunction t = cFunction (getName t) (map toExtendedTerm' (getChildren t))

instance (TermTypeBuilder tt) => PositionTerm (ExtendedTermT tt) StandardTreePath where
    makeContextPair t = (replaceSubtree t cHole, extractNode t)
    makeMonadicTerm t = path2MonadicTreeC (getRoot t) (getPath t)

instance  (Term t1) => ConvertableTerm t1 ExtendedTerm where
    convertTerm = toExtendedTerm

