module Gersdorf.ContextMatching.Base.SimpleSubstitution (
   SimpleSubstitutionAlt,
   module Gersdorf.ContextMatching.Base.Substitution
) where

import Data.Maybe

import Gersdorf.ContextMatching.Base.HeapTerm
import Gersdorf.ContextMatching.Base.Substitution hiding (SubstitutionSet)

newtype (PositionTerm t tp) => SubstitutionSet t tp = SubstitutionSet [SubstitutionAtom t]
newtype (PositionTerm t tp) => SubstitutionAlt t tp = SubstitutionAlt [SubstitutionSet t tp]

type SimpleSubstitutionSet = SubstitutionSet HeapTerm HeapPath
type SimpleSubstitutionAlt = SubstitutionAlt HeapTerm HeapPath

instance (PositionTerm t tp) => BotToppable (SubstitutionSet t tp) where
    isTop (SubstitutionSet substs) = all isTop substs
    isBot (SubstitutionSet substs) = any isBot substs

    isFlatTop (SubstitutionSet [s]) = isFlatTop s
    isFlatTop _ = False

    isFlatBot (SubstitutionSet [s]) = isFlatBot s
    isFlatBot _ = False

instance (PositionTerm t tp) => BotToppable (SubstitutionAlt t tp) where
    isTop (SubstitutionAlt alts) = all isTop alts
    isBot (SubstitutionAlt alts) = all isBot alts

    isFlatTop (SubstitutionAlt [a]) = isFlatTop a
    isFlatTop _ = False
    isFlatBot (SubstitutionAlt [a]) = isFlatBot a
    isFlatBot _ = False

instance SubstitutionBuilder (SubstitutionAlt HeapTerm HeapPath) HeapTerm HeapPath where
    substTop = substitutionAlt1 SubstTop
    substBot = substitutionAlt1 SubstBot

    substVar var target = 
        substitutionAlt1 (SubstVar (getType var) target)
    substContextVar cvar target holePos = 
        substitutionAlt1 (SubstContextVar (getType cvar) (fromJust $ makeContextAdress target holePos))

    substAnd [] = substitutionAlt1 SubstTop
    substAnd [s] = s
    substAnd (s:ss) = substAnd2 s (substAnd ss)

    substOr [] = substitutionAlt1 SubstTop
    substOr [s] = s
    substOr (s:ss) = substOr2 s (substOr ss)

instance SubstitutionGenerator (SubstitutionAlt HeapTerm HeapPath) HeapTerm where
    generate s = toList s
        where toList (SubstitutionAlt alts) = map (\(SubstitutionSet set) -> Substitution set) alts
    simplify = simplifySubstitutionAlt 

instance SubstitutionGenerator (SubstitutionSet HeapTerm HeapPath) HeapTerm where
    generate (SubstitutionSet set) = [Substitution set]
    simplify = simplifySubstitutionSet

substitutionAlt1 :: SubstitutionAtom HeapTerm -> SimpleSubstitutionAlt
substitutionAlt1 s = SubstitutionAlt [SubstitutionSet [s]]

substAnd2 :: SimpleSubstitutionAlt -> SimpleSubstitutionAlt -> SimpleSubstitutionAlt
substAnd2 (SubstitutionAlt dss1) (SubstitutionAlt dss2) =
        SubstitutionAlt [ SubstitutionSet (css1 ++ css2) | (SubstitutionSet css1) <- dss1, (SubstitutionSet css2) <- dss2 ]

substOr2 :: SimpleSubstitutionAlt -> SimpleSubstitutionAlt -> SimpleSubstitutionAlt
substOr2 (SubstitutionAlt dss1) (SubstitutionAlt dss2) = SubstitutionAlt (dss1 ++ dss2)

-- |Wir filtern nur nach Alternativen, die Bot sind. Dies kann lazy
-- geschehen, da nur solange iteriert werden muss, bis eine non-bot
-- Alternative gefunden wird.
-- Die berprfung, ob die Gesamtheit der Alternativen Top ist,
-- wrde die Lazyness zerstren, da in diesem Falle die Frage lautet
-- "gibt es unter allen Alternativen mindestens eine, die Top ist?".
-- In diesem Falle mssten im worst case alle Alternativen untersucht
-- und damit auch berechnet werden.
simplifySubstitutionAlt :: SimpleSubstitutionAlt -> SimpleSubstitutionAlt
simplifySubstitutionAlt (SubstitutionAlt dss) = 
    SubstitutionAlt $ filter (not . isBot . simplifySubstitutionSet) dss

-- |Wir filtern ein Substitutionset zunchst danach, ob es Bot ist.
-- Dafr mssen alle Klauseln berechnet werden. Dies ist jedoch auch notwendig
-- fr die Ausgabe der Klauseln und damit vertretbar.
simplifySubstitutionSet :: SimpleSubstitutionSet -> SimpleSubstitutionSet
simplifySubstitutionSet (SubstitutionSet substset) = 
    removeBotSubst $ SubstitutionSet (filter (not . isTop) substset)
    where removeBotSubst :: SimpleSubstitutionSet -> SimpleSubstitutionSet
          removeBotSubst ss = if (isBot ss) then SubstitutionSet [SubstBot]
                                            else ss

