-- |Definition einer Typklasse fr Substitutionen
module Gersdorf.ContextMatching.Base.Substitution (
    SubstitutionBuilder(..),
    SubstitutionAtom(..),
    Substitution(..),
    SubstitutionGenerator(..),
    Substitutable(..),
    SubstitutionSet(..),
    applySubstitutionGenerator,
    BotToppable(..)
 ) where

import Gersdorf.ContextMatching.Base.PositionTerm
import Gersdorf.ContextMatching.Base.Utilities
import Gersdorf.ContextMatching.Base.ExtendedTerm
import Gersdorf.Data.TreeFiniteMap 
import Debug.Trace

--
-- KLASSEN
--

-- |Ein Datum, das BotToppable ist, kann auf Bot oder Top berprft werden.
-- Semantik von Bot/Top:
--   * Bot bedeutet, dass diese Sublsung die Gesamtlsung unmglich macht
--   * Top bedeutet, dass diese Sublsung keinen Anteil an der Gesamtlsung
--     hat und daher weggelassen werden kann.
class BotToppable t where
    isTop :: t -> Bool
    isBot :: t -> Bool
    isFlatTop :: t -> Bool
    isFlatBot :: t -> Bool
    isBotTop :: t -> Bool
    isBotTop t = isTop t || isBot t
    isFlatBotTop :: t -> Bool
    isFlatBotTop t = isFlatTop t || isFlatBot t

-- |Ein SubstitutionGenerator kann Substitutionen aus sich heraus generieren.
-- Er generiert dazu eine Liste von Substitutionen, die als alternative
-- Lsungen angesehen werden knnen.
class (Term t) => SubstitutionGenerator sg t | sg -> t where
   simplify :: sg -> sg 
   simplify = id
   generate :: sg -> [Substitution t]

-- |Ein SubstitutionBuilder ist ein Datentyp, mit dessen Hilfe eine Substitution
-- zusammengebaut werden kann und der daher die genannten elementaren Operationen
-- untersttzt.
-- Der SubstitutionBuilder ist parameterisiert:
--   (*) t - der TermTyp, mit dessen Hilfe die Substitution aufgebaut wird
--   (*) p - Der Typ der Positionen, die innerhalb des Termtypen verwendet werden

class (PositionTerm t tp,  
       SubstitutionGenerator s t)  => SubstitutionBuilder s t tp where
   substTop :: s
   substBot :: s  
   substVar :: t -> t -> s  
   substContextVar :: t -> t -> Int -> s  

   substAnd :: [s] -> s  
   substOr :: [s] -> s  

-- |Ein Term ist dann ein Substitutable, wenn auf ihm eine Substitution angewendet
-- werden kann. Dafr muss er die genannte Operation untersttzen.
class Substitutable t where
    applySubstitution :: (Term t2) => t -> Substitution t2 -> (Substitution t2, t)
    isCorrectUnderSubst :: (Term t2) => t -> Substitution t2 -> Bool
    isCorrectUnderSubst _ _ = True

class (Term t) => SubstitutionSet s t where
    emptySubstitution :: s t

    lookupSubstitution :: s t -> TermType -> Maybe t 

    hasSubstitution :: s t  -> TermType -> Bool
    hasSubstitution s tt = 
        case (lookupSubstitution s tt) of
            Nothing -> False
            (Just _) -> True
    
    lookupSubstitutionST :: s t -> TermType -> Maybe ExtendedTerm 
    lookupSubstitutionST s tt = 
        do substitutionTerm <- lookupSubstitution s tt
           return (toExtendedTerm substitutionTerm)
        
    bindSubstitution :: s t  -> TermType -> t  -> s t 

--
-- Datentypen
--

-- |Ein SubstitutionsAtom ist eine einzelne Substitution fr eine
-- Variable oder ContextVariable oder aber Bot oder Top.
-- Aus einzelnen Substitutionsatomen werden Substitutionen zusammegesetzt.
-- (siehe Substitution).
data Term t => SubstitutionAtom t = SubstBot | SubstTop
                                    | SubstVar TermType t
                                    | SubstContextVar TermType t
    deriving Eq

-- |Eine Substitution ist eine Menge von Substitutionsatomen, die zusammen auf
-- einem Term angewendet werden sollen. Substitutionen sind parameterisiert ber
-- Termen, da bei Variablen und Kontextvariablen Terme ersetzt werden.
newtype (Term t) => Substitution t = Substitution [SubstitutionAtom t]

--
-- INSTANZEN
--

instance (Term t) => SubstitutionSet Substitution t where
    emptySubstitution = Substitution []

    lookupSubstitution (Substitution sas) tt = lookupSubstitution' sas tt
        where
        lookupSubstitution' [] _ = Nothing
        lookupSubstitution' ((SubstVar tt t):sas) tt2 | tt == tt2 = Just t
        lookupSubstitution' ((SubstContextVar tt t):sas) tt2 | tt == tt2 = Just t
        lookupSubstitution' (s:sas) tt2 = lookupSubstitution' sas tt2

    bindSubstitution (Substitution sas) t@(TermVariable _) tt = Substitution ((SubstVar t tt):sas)
    bindSubstitution (Substitution sas) t@(TermContextVariable _) tt = Substitution ((SubstContextVar t tt):sas)
    bindSubstitution s _ _ = s

    
-- |Substitutionsatome sind BotToppable, da sie entweder Bot oder Top sein knnen.
instance (Term t) => BotToppable (SubstitutionAtom t) where
    isTop SubstTop = True
    isTop _ = False

    isBot SubstBot = True
    isBot _ = False

    isFlatTop = isTop
    isFlatBot = isTop

instance (Term t) => BotToppable (Substitution t) where
    -- Da die Substitution eine UND-Verknpfung ist, ist jede Substitution, in
    -- der ein Bot enthalten ist, ungltig.
    isBot (Substitution ss) = any isBot ss
    isTop (Substitution ss) = all isTop ss

    isFlatBot (Substitution [SubstBot]) = True
    isFlatBot _ = False
    isFlatTop (Substitution [SubstTop]) = True
    isFlatTop _ = False

-- |Ein SubstitutionsAtom ist selbst ein SubstitutionGenerator, indem es
-- sich selbst als Substitution generiert.
instance (Term t) => SubstitutionGenerator (SubstitutionAtom t) t where
    simplify = id
    generate s = [Substitution [s]]

instance Substitutable ExtendedTerm where
      applySubstitution t substset = (substset, transformTree (extractNode' fmapSet) t)
           where fmapSet = toFMapSubstitution substset

                 extractNode' smap t = 
                    case (lookupSubstitution' smap t) of
                        Nothing  -> TANothing
                        (Just r) -> TACompleteTree (extractNode r)

                 lookupSubstitution' :: (Term t) => FMapSubstitution t -> ExtendedTerm-> Maybe ExtendedTerm
                 lookupSubstitution' smap lt | isVariable lt = lookupSubstitutionST smap (TermVariable $ getName lt) 
                                             | isContextVariable lt  = 
                   do ct <- lookupSubstitutionST smap (TermContextVariable $ getName lt) 
                      let ctxInset = extractNode $ head $ getChildren lt
                      instantiateContext ct ctxInset
                                             | otherwise = Nothing

                 applyContext :: ExtendedTerm -> ExtendedTerm -> ExtendedTerm
                 applyContext c it = transformTree applyContext' c
                   where applyContext' t | isHole t = TACompleteTree (extractNode it)
                                         | otherwise = TANothing

instance (Term t) => Show (SubstitutionAtom t) where
    showsPrec _ (SubstBot) r = showString "<bot>" r
    showsPrec _ (SubstTop) r = showString "<top>" r
    showsPrec _ (SubstVar var t) r = showString "<" (shows t (showString "/" (shows var (showString ">" r))))
    showsPrec _ (SubstContextVar cvar t) r = showString "<" (shows t (showString ("/" ++ (show cvar) ++ ">") r))

instance (Term t) => Show (Substitution t) where
    showsPrec _ (Substitution ss) r = showPredList "{ " ss ", " " }" r

--
-- INTERNA
--

newtype (Term t) => FMapSubstitution t = 
    FMapSubstitution (TreeFiniteMap TermType (SubstitutionAtom t))

instance (Term t) => SubstitutionSet FMapSubstitution t where
    emptySubstitution = FMapSubstitution emptyFM

    lookupSubstitution (FMapSubstitution fsub) t = case (lookupFM t fsub) of
                                   Just (SubstVar _ tt) -> Just tt
                                   Just (SubstContextVar _ tt) -> Just tt
                                   _ -> Nothing

    bindSubstitution (FMapSubstitution fsub) t@(TermVariable _) tt = 
        FMapSubstitution $ bindFM t (SubstVar t tt) fsub
    bindSubstitution (FMapSubstitution fsub) t@(TermContextVariable _) tt = 
        FMapSubstitution $ bindFM t (SubstContextVar t tt) fsub
    bindSubstitution f _ _ = f

toFMapSubstitution :: (Term t) => Substitution t -> FMapSubstitution t
toFMapSubstitution (Substitution ss) = FMapSubstitution $ listToFM 
                        $ map convert2Pairs $ filter (not . isBotTop) ss
    where
      convert2Pairs :: (Term t) => SubstitutionAtom t -> (TermType, SubstitutionAtom t)
      convert2Pairs s@(SubstVar t _) = (TermVariable (getName t), s)
      convert2Pairs s@(SubstContextVar t _)  = (TermContextVariable (getName t), s)
          
applySubstitutionGenerator :: (SubstitutionGenerator sg sgt, Substitutable t) => 
   sg -> t -> [(Substitution sgt, t)]
applySubstitutionGenerator sg t = map (applySubstitution t) (generate sg)


