-- |This module contains declarations needed to work with Terms,
-- especially a class defining all methods allowed on Terms

module Gersdorf.ContextMatching.Base.Term (
    Term(..),
    AnalysisTerm(..),
    TermBuilder(..),
    SubTermBuilder(..),
    TermIterator(..),
    TermType(..),
    TermTypeable(..),
    TermTypeBuilder(..),
    ConvertableTerm(..),
    Linearizable(..),
    showTerm,
    TermNode,
    module Gersdorf.Data.Tree
) where

import Debug.Trace

import Gersdorf.Data.Tree
import Gersdorf.Data.List
import Gersdorf.ContextMatching.Base.Utilities

data TermType = TermHole
                | TermVariable String
                | TermContextVariable String 
                | TermFunction String
    deriving (Eq, Ord)

instance Show TermType where
    showsPrec _ tt r = showString (getName tt) r

class (Ord t) => TermTypeable t where
    getName :: t -> String
    getName = getName . getType

    -- |Tester Function for Variables
    isVariable :: t -> Bool
    isContextVariable :: t -> Bool
    isHole :: t -> Bool
    isFunction :: t -> Bool
    isVariable t = case (getType t) of
                        TermVariable _ -> True
                        _ -> False
    isContextVariable t = case (getType t) of
                               TermContextVariable _ -> True
                               _ -> False
    isHole t = case (getType t) of
                    TermHole -> True
                    _ -> False
    isFunction t = case (getType t) of
                        TermFunction _ -> True
                        _ -> False


    -- |Get the term Type as defined above
    getType :: t -> TermType
 
instance TermTypeable TermType where
    getName TermHole                = "_"
    getName (TermVariable n)        = n
    getName (TermContextVariable n) = n
    getName (TermFunction n)        = n

    getType = id

class (TermTypeable tt) => TermTypeBuilder tt where
    cFunctionType :: String -> tt
    cHoleType :: tt
    cVarType :: String -> tt
    cContextVarType :: String -> tt

instance TermTypeBuilder TermType where
    cFunctionType n = TermFunction n
    cHoleType = TermHole
    cVarType n = TermVariable n
    cContextVarType n = TermContextVariable n

class Linearizable l where
    linearize :: Int -> l -> (Int, l)

-- |A Type class representing all those data types giving
-- access to /terms/
class (Ord t, Linearizable t, NavigatableTree t, Show t, TermIterator t, TermTypeable t) => Term t where
    isSolvedVariable :: t -> Bool
    isSolvedVariable t = (isVariable t) || (isContextVariable t && isHole (head $ getChildren t))

    isGround :: t -> Bool
    isGround = not . any (\t -> isVariable t || isContextVariable t) . toList

    -- |True, falls dieser Term ein Kontext ist, d.h. er enthlt ein Loch
    isContext :: t -> Bool
    isContext = any (\t -> isHole t) . toList

    -- |Get count of holes in this term
    countHoles :: t -> Int
    countHoles = fold (\c t -> if (isHole t) then (c+1) else c) 0

    getHolesDeep :: t -> [t]
    getHolesDeep t | isHole t  = [t]
                   | otherwise = concat $ map getHolesDeep (getChildren t)

    getHolesFlat :: t -> [t]
    getHolesFlat = filter isHole . getChildren

    holeSignatureFlat :: t -> [Int]
    holeSignatureFlat = map (\t -> if (isHole t) then 1 else 0) . getChildren

    holeSignatureDeep :: t -> [Int]
    holeSignatureDeep = map countHoles . getChildren

    -- | holt alle Konstanten (Funktionen, die Bltter sind), die nur einmal
    --   im Term vorkommen.
    getUniqueConstants :: t -> [t]
    getUniqueConstants = uniques . filter isFunction . leaves

   

class (Listable t) => TermIterator t where
    -- |accumulates sth, i. e. iterates over the Term of type @t@,
    -- computes a value of type @a@ on every node using the function of
    -- type @t->a@ and combines those values using the second function
    -- of type @a->a@ to compute a total result.
    fold :: (a->t->a)  -- ^The function accumulating a result value of type a and a term
      -> a             -- ^The initial value for accumulation
      -> t             -- ^The term to be accumulated
      -> a             -- ^The result value
    fold accum init = foldl accum init . toList

type TermNode ct = (Maybe (ct, Int))     -- ^ father and position of this child
    -> ct                                -- ^ created Term (Tree t over a)

class SubTermBuilder ct where
    -- |factory function to create a Function
    cFunction :: String -- ^Name of the Function
        -> [TermNode ct] -- ^List of Argument terms
        -> TermNode ct  -- ^resulting Function term
    -- |factory function to create a Variable
    cVar :: String      -- ^Name of Variable
        -> TermNode ct  -- ^Resulting Variable term
    -- |factory function to create a ContextVariable
    cContextVar :: String  -- ^Name of Context Variable
        -> TermNode ct    -- ^Term that instantiates the Context
        -> TermNode ct    -- ^Resulting ContextVar term
    -- |factory function to create a Hole
    cHole :: TermNode ct   -- ^Resulting Hole Term

    cTermNode :: TermType
        -> [TermNode ct]
        -> TermNode ct
    cTermNode TermHole _ = cHole
    cTermNode (TermVariable n) _ = cVar n
    cTermNode (TermContextVariable n) chs = cContextVar n (head chs)
    cTermNode (TermFunction n) chs = cFunction n chs

class (Term t, SubTermBuilder t) => TermBuilder t  where
    cFunctionT :: String
        -> [t]
        -> TermNode t 

    cContextVarT :: String
        -> t 
        -> TermNode t

    cTerm :: TermNode t -> t
 
    replaceTerm :: t       -- ^ (Sub-)Term, der ersetzt werden soll
        -> TermNode t     -- ^ neuer TermNode fr den alten Subterm
        -> t               -- ^ Wurzel des neuen GesamtTerms

    -- | extrahiert einen Subterm, so dass dieser Subterm unabhngig von
    --   seinen Vatertermen ist (d. h. die Subtermwurzel wird zur Termwurzel)
    extractSubTerm :: t      
        -> t

class (Term t) => AnalysisTerm t where
    getHoleCount :: t -> Int
    getLeftHoles :: t -> Int
    getRightHoles :: t -> Int


class (Term t1, Term t2) => ConvertableTerm t1 t2 where
    convertTerm :: t1 -> t2 

{-class (Term t1, ContainerTree 
    toTree :: (ContainerTree tr TermType) => t -> tr TermType
-}
-- Wir knnen leider nicht einfach schreiben
--    instance (Term t) => Show t where ...
-- da wir damit in undecidable instances laufen und -fallow-undecidable-instances
-- nutzen mssten. Deshalb eine etwas umstndlichere Formulierung.
showTerm :: (Term t) => t -> ShowS
showTerm t r = showPredList (getName t ++ getOpener tType) (getChildren t) "," (getCloser tType) r
        where
        tType :: TermType
        tType = getType t

        getOpener :: TermType -> String
        getOpener (TermFunction _) = "("
        getOpener (TermContextVariable _) = "["
        getOpener _ = ""

        getCloser :: TermType -> String
        getCloser (TermFunction _) = ")"
        getCloser (TermContextVariable _) = "]"
        getCloser _ = ""
 

