-- |This module contains declarations needed to work with Terms,
-- especially a class defining all methods allowed on Terms

module Gersdorf.ContextMatching.Base.Term (
    Term(..),
    TermBuilder(..),
    SubTermBuilder(..),
    TermIterator(..),
    TermType(..),
    TermTypeable(..),
    ConvertableTerm(..),
    showTerm,
    TermNode,
    module Gersdorf.Data.Tree
) where

import Gersdorf.Data.Tree
import Gersdorf.ContextMatching.Base.Utilities

data TermType = TermHole
                | TermVariable String
                | TermContextVariable String 
                | TermFunction String
    deriving (Eq, Ord)

instance Show TermType where
    showsPrec _ tt r = showString (getName tt) r

class (Eq t) => TermTypeable t where
    getName :: t -> String
    getName = getName . getType

    -- |Tester Function for Variables
    isVariable :: t -> Bool
    isContextVariable :: t -> Bool
    isHole :: t -> Bool
    isFunction :: t -> Bool
    isVariable t = case (getType t) of
                        TermVariable _ -> True
                        _ -> False
    isContextVariable t = case (getType t) of
                               TermContextVariable _ -> True
                               _ -> False
    isHole t = case (getType t) of
                    TermHole -> True
                    _ -> False
    isFunction t = case (getType t) of
                        TermFunction _ -> True
                        _ -> False


    -- |Get the term Type as defined above
    getType :: t -> TermType
 
instance TermTypeable TermType where
    getName TermHole                = "_"
    getName (TermVariable n)        = n
    getName (TermContextVariable n) = n
    getName (TermFunction n)        = n

    getType = id

-- |A Type class representing all those data types giving
-- access to /terms/
class (ContainerTree t e, Show (t e), Eq (t e), TermIterator (t e), TermTypeable (t e), TermTypeable e) => Term t e where
    isSolvedVariable :: t e -> Bool
    isSolvedVariable t = (isVariable t) || (isContextVariable t && isHole (head $ getChildren t))

    isGround :: t e -> Bool
    isGround = not . any (\t -> isVariable t || isContextVariable t) . toList

    -- |True, falls dieser Term ein Kontext ist, d.h. er enthlt ein Loch
    isContext :: t e -> Bool
    isContext = any (\t -> isHole t) . toList

    -- |Get count of holes in this term
    countHoles :: t e -> Int
    countHoles = fold (\c t -> if (isHole t) then (c+1) else c) 0 

class (Listable t) => TermIterator t where
    -- |accumulates sth, i. e. iterates over the Term of type @t@,
    -- computes a value of type @a@ on every node using the function of
    -- type @t->a@ and combines those values using the second function
    -- of type @a->a@ to compute a total result.
    fold :: (a->t->a)  -- ^The function accumulating a result value of type a and a term
      -> a             -- ^The initial value for accumulation
      -> t             -- ^The term to be accumulated
      -> a             -- ^The result value
    fold accum init = foldl accum init . toList

type TermNode ct tt = (Maybe (ct tt, Int))     -- ^ father and position of this child
    -> ct tt                                   -- ^ created Term (Tree t over a)

class (TermTypeable tt) => SubTermBuilder ct tt where
    -- |factory function to create a Function
    cFunction :: String -- ^Name of the Function
        -> [TermNode ct tt] -- ^List of Argument terms
        -> TermNode ct tt  -- ^resulting Function term
    -- |factory function to create a Variable
    cVar :: String      -- ^Name of Variable
        -> TermNode ct tt -- ^Resulting Variable term
    -- |factory function to create a ContextVariable
    cContextVar :: String  -- ^Name of Context Variable
        -> TermNode ct tt   -- ^Term that instantiates the Context
        -> TermNode ct tt   -- ^Resulting ContextVar term
    -- |factory function to create a Hole
    cHole :: TermNode ct tt  -- ^Resulting Hole Term

    cTermNode :: TermType
        -> [TermNode ct tt]
        -> TermNode ct tt
    cTermNode TermHole _ = cHole
    cTermNode (TermVariable n) _ = cVar n
    cTermNode (TermContextVariable n) chs = cContextVar n (head chs)
    cTermNode (TermFunction n) chs = cFunction n chs


class (Term t tt, SubTermBuilder ct tt) => TermBuilder t tt ct where
    cFunctionT :: String
        -> [t tt]
        -> TermNode ct tt

    cContextVarT :: String
        -> t tt
        -> TermNode ct tt

    cTerm :: TermNode ct tt -> t tt
 
    replaceTerm :: t tt      -- ^ (Sub-)Term, der ersetzt werden soll
        -> TermNode ct tt    -- ^ neuer TermNode fr den alten Subterm
        -> t tt              -- ^ Wurzel des neuen GesamtTerms

class (ConvertableContainerTree t1 t2 tt, Term t1 tt, Term t2 tt) => ConvertableTerm t1 t2 tt where
    convertTerm :: t1 tt -> t2 tt

{-class (Term t1, ContainerTree 
    toTree :: (ContainerTree tr TermType) => t -> tr TermType
-}
-- Wir knnen leider nicht einfach schreiben
--    instance (Term t) => Show t where ...
-- da wir damit in undecidable instances laufen und -fallow-undecidable-instances
-- nutzen mssten. Deshalb eine etwas umstndlichere Formulierung.
showTerm :: (Term t tt) => t tt -> ShowS
showTerm t r = showPredList (getName t ++ getOpener tType) (getChildren t) "," (getCloser tType) r
        where
        tType :: TermType
        tType = getType t

        getOpener :: TermType -> String
        getOpener (TermFunction _) = "("
        getOpener (TermContextVariable _) = "["
        getOpener _ = ""

        getCloser :: TermType -> String
        getCloser (TermFunction _) = ")"
        getCloser (TermContextVariable _) = "]"
        getCloser _ = ""
 

