module Gersdorf.ContextMatching.Base.TermGenerator (
    unsafeGenerateUP,
    unsafeGenerateUPn,

    generateUP,
    generateUPn,
    generateTermEq,

    TermGeneratorParameter(..),
    TermTypeParameter(..)
) where

import Control.Monad.State
import Data.Array
import Data.Maybe

import Gersdorf.Algorithms.Random
import Gersdorf.Data.TreeFiniteMap
import Gersdorf.Monad.IOUtil

import Gersdorf.ContextMatching.Base.ExtendedTerm
import Gersdorf.ContextMatching.Base.Substitution
import Gersdorf.ContextMatching.Base.SimpleUnificationProblem
import Gersdorf.ContextMatching.Base.SimpleTermEquation
import Gersdorf.ContextMatching.Base.Utilities

import System.IO.Unsafe (unsafePerformIO)

-------------------------------------------------------------------------
-- Hauptfunktionen (Fassade)
-------------------------------------------------------------------------

unsafeGenerateUP :: 
    TermGeneratorParameter 
    -> (ListEquationSet SimpleTermEquation ExtendedTerm)
unsafeGenerateUP p = unsafePerformIO $ generateUP p

unsafeGenerateUPn :: 
    TermGeneratorParameter 
    -> Int 
    -> [ListEquationSet SimpleTermEquation ExtendedTerm]
unsafeGenerateUPn p n = unsafePerformIO $ generateUPn p n

generateUPn :: 
    TermGeneratorParameter 
    -> Int 
    -> IO [ListEquationSet SimpleTermEquation ExtendedTerm]
generateUPn p n =
    sequenceG $ replicate n (generateUP p)

generateUP :: 
    TermGeneratorParameter 
    -> IO (ListEquationSet SimpleTermEquation ExtendedTerm)
generateUP p =
    do eq <- generateTermEq p
       return (cProblem [eq])

generateTermEq :: 
    TermGeneratorParameter 
    -> IO (SimpleTermEquation ExtendedTerm)
generateTermEq p = 
    do feed           <- createFeed p 
       let startState =  TermState { tsFeed = feed, occurenceVar = 0, occurenceCVar = 0 }
       let action     =  generateTermSt (termSize p) (variableSize p)
       (t1, t2)       <- evalStateT action startState
       return (cEquation t1 t2)

type Weight = Int

data TermTypeParameter = TermTypeParameter {
    termTypeCount :: Int,
    termTypeWeight :: Weight
}

instance Show TermTypeParameter where
    showsPrec _ ttp r =
        showRecordHelper "TermTypeParameter"
            [("count", show . termTypeCount),
             ("weight", show . termTypeWeight)] ttp r

data TermGeneratorParameter = TermGeneratorParameter {
    termSize :: Int,
    variableSize :: Int,   -- assertion: variableSize < termSize
    varParameter :: TermTypeParameter,
    cvarParameter :: TermTypeParameter,
    funcParameter :: TermTypeParameter,
    holeParameter :: TermTypeParameter,
    maxOccurenceVar :: Int,
    maxOccurenceCVar :: Int,
    maxArity :: Int,
    randomRangeF :: (Int, Int) -> IO Int,
    distributionF :: Int -> Int -> IO [Int]
}

instance Show TermGeneratorParameter where
    showsPrec _ tgp r =
        showRecordHelper "TermGeneratorParameter"
            [("termSize", show . termSize),
             ("variableSize", show . variableSize),
             ("varParameter", show . varParameter),
             ("cvarParameter", show . cvarParameter),
             ("funcParameter", show . funcParameter),
             ("holeParameter", show . holeParameter),
             ("maxOccurenceVar", show . maxOccurenceVar),
             ("maxOccurenceCVar", show . maxOccurenceCVar),
             ("maxArity", show . maxArity)] tgp r

-----------------------------------------------------------------------------    
-- eigentliche Generatorroutine.
-----------------------------------------------------------------------------

generateTermSt :: 
    Int      -- termSize
    -> Int   -- variableSize
    -> GeneratorState (ExtendedTerm, ExtendedTerm)
generateTermSt ts vs = 
    do l       <- generateTerm vs
       let lt  =  cTree l

       varNum  <- getParameter feedNumVar
       cvarNum <- getParameter feedNumCVar

       let delta            =  ts - vs
           varDistribution  =  numberVariables lt varNum
           nvars            =  length varDistribution
           cvarDistribution =  numberCVariables lt cvarNum
           bothDistribution =  varDistribution ++ cvarDistribution

       if (null bothDistribution)
           then return (lt, lt)     -- keine Variablen => es kann kein Term untergebracht werden!
           
           else do let varTermTypes     =  map (\(i, _) -> getVariableTermType i) varDistribution
                       cvarTermTypes    =  map (\(i, _) -> getContextVariableTermType i) cvarDistribution
                       weights          =  map snd bothDistribution

                   randomDistributed    <- lift $ distributeWeightedIO delta weights
                   --debug ("randomDistributed: " ++ show randomDistributed)
       
                   let varRandoms       =  zip varTermTypes (take nvars randomDistributed)
                       cvarRandoms      =  zip cvarTermTypes (drop nvars randomDistributed)

                   varT  <- generateSubstitutedTerm lt substituteVar varRandoms 
                   cvarT <- generateSubstitutedTerm varT substituteCVar cvarRandoms 
               
                   return (lt, cvarT)

generateSubstitutedTerm :: (Term t, Substitutable t) =>
    t
    -> ((TermType, Int) -> GeneratorState (SubstitutionAtom t))
    -> [(TermType, Int)]      -- TermTypes mit gre der zu generierenden Substitutions
    -> GeneratorState t
generateSubstitutedTerm t substF tts =
    do substAtoms <- mapM substF tts
       let subst  =  Substitution substAtoms
       return $ snd $ (applySubstitution t subst)

substituteVar :: 
    (TermType, Int)
    -> GeneratorState (SubstitutionAtom ExtendedTerm)
substituteVar (tt, n) =
    do groundTermNode <- generateGroundTerm n
       let groundTerm = cTerm groundTermNode
           substAtom  = SubstVar tt groundTerm
       return substAtom
    
substituteCVar :: 
    (TermType, Int)
    -> GeneratorState (SubstitutionAtom ExtendedTerm)
substituteCVar ttDescr@(tt, _) =
    do (SubstVar _ groundTerm) <- substituteVar ttDescr
       -- hier jetzt zufllig ein Blatt auswhlen!
       holeNode                <- lift $ pickFromIO $ leaves groundTerm
       let contextTerm         =  makeContext holeNode
           substAtom           =  SubstContextVar tt contextTerm
       return substAtom
    
generateGroundTerm :: Int -> GeneratorState ExtendedTermNode
generateGroundTerm r
    | r == 1 = do rt                <- randomTerm (0,0) [feedFunctionF]
                  cRNode rt []
    | r > 1  = do rt@(_, ar)        <- randomTerm (1, r-1) [feedFunctionF]
                  df                <- getParameter feedDistribute
                  arityDistribution <- lift $ df (r-1) ar
                  cRNode rt (map generateGroundTerm arityDistribution)
    
generateTerm :: Int -> GeneratorState ExtendedTermNode
generateTerm r  
    | r == 1 = do rt                <- randomTerm (0,0) [feedVariableF, feedHoleF, feedFunctionF]
                  cRNode rt []
    | r > 1  = do rt@(_, ar)        <- randomTerm (1, r-1) [feedFunctionF, feedCVariableF]
                  df                <- getParameter feedDistribute
                  arityDistribution <- lift $ df (r-1) ar
                  cRNode rt (map generateTerm arityDistribution)

randomTerm :: (Int, Int) -> [TermFeed -> FeedFunction] -> GeneratorState RTermType
randomTerm r fs =
    do ft               <- getParameter feedTermF
       selected@(tt, _) <- ft r fs

       updateOccurence tt
       return selected

updateOccurence :: TermType -> GeneratorState ()
updateOccurence tt =
    if (isVariable tt)
        then modify updateOccurenceVar
        else if (isContextVariable tt)
                then modify updateOccurenceCVar
                else return ()
    where
    updateOccurenceVar :: TermState -> TermState
    updateOccurenceVar s = s { occurenceVar = (occurenceVar s) + 1 }
    updateOccurenceCVar :: TermState -> TermState
    updateOccurenceCVar s = s { occurenceCVar = (occurenceCVar s) + 1 }
 
-----------------------------------------------------------------------------
-- Statemonadfunktionen
-----------------------------------------------------------------------------

-- | Interner Status des Termgenerierers.
data TermState = TermState {
    tsFeed :: TermFeed,    -- ^ feed, der den Generierer mit Subtermen versorgt
    occurenceVar :: Int,   -- ^ Anzahl bereits generierter Variablen
    occurenceCVar :: Int   -- ^ Anzahl bereits generierter Kontextvariablen
}

-- | StateMonade des Termgenerierers. r ist das Resultat.
--   Ist ein StateTransformer, da sowohl die IO- als auch eine Statemonade benutzt wird.
type GeneratorState r = StateT TermState IO r

-- | Eine FeedFunction nimmt ein Paar aus minimaler und maximaler Stelligkeit
--   und gibt ggf. (falls diese Anforderung erfllt werden kann) ein Paar
--   aus Gewicht und RTermType zurck
type FeedFunction = 
    (Int, Int)                                    -- ^ (minimale, maximale) Stelligkeit
    -> GeneratorState (Maybe (Weight, RTermType)) -- ^ (Gewicht, TermTyp incl. Stelligkeit)

data TermFeed = TermFeed {
    feedVariableF  :: FeedFunction,     -- ^ feed, der mit Variablen versorgt
    feedCVariableF :: FeedFunction,     -- ^ feed, der mit Kontextvariablen versorgt
    feedFunctionF  :: FeedFunction,     -- ^ feed, der mit Funktionen versorgt
    feedHoleF      :: FeedFunction,     -- ^ feed, der mit Lchern versorgt

    feedNumVar     :: Int,              -- ^ Anzahl zur Verfgung stehender Variablen
    feedNumCVar    :: Int,              -- ^ Anzahl zur Verfgung stehender Kontextvariablen

    -- | Funktion, die mit Termen versorgt
    feedTermF      :: (Int, Int)        -- ^ gewnschte Stelligkeit (min, max)
        -> [TermFeed -> FeedFunction]   -- ^ feeds fr Terme, die gewnscht werden
        -> GeneratorState RTermType,    -- ^ gewhlter Subterm

    -- ^ Zufallsfunktion, liefert Wert zwischen (min, max)
    feedRange      :: (Int, Int)        -- ^ (min, max)
        -> IO Int,                      -- ^ Zufallswert
    -- ^ Zufallsfunktion, belegt n Pltze mit r Werten
    feedDistribute :: Int               -- ^ r
        -> Int                          -- ^ n
        -> IO [Int]                     -- Verteilte Werte mit n Elementen
}

-- | Funktion, die aus dem TermGeneratorParameter (Datentyp zur externen Parameterisierung)
--   einen Termfeed berechnet und dabei die feeds generiert.
createFeed :: 
    TermGeneratorParameter  -- ^ externer Parameter fr den TermGenerator
    -> IO TermFeed          -- ^ Feed, der fr den Termgenerator intern benutzt wird.
createFeed p =
    do functionTable <- createFunctionTable p
       let feed = TermFeed { 
                    feedVariableF  = feedVariable p,
                    feedCVariableF = feedCVariable p,
                    feedFunctionF  = feedFunction p functionTable,
                    feedHoleF      = feedHole p,
                    feedTermF      = feedTerm,
                    feedDistribute = distributionF p,
                            feedRange      = randomRangeF p,

                            feedNumVar     = termTypeCount (varParameter p),
                            feedNumCVar    = termTypeCount (cvarParameter p)
                          }
       return feed
               
-- | Funktion, die abhngig von der Parameterisierung eine zufllige Variable
--   liefert. Dabei wird die aktuelle Anzahl bereits generierter und das
--   diesbezgliche Limit von Variablen bercksichtigt.
feedVariable :: TermGeneratorParameter -> FeedFunction
feedVariable p (min, max) = 
    do currentVars <- gets occurenceVar
       let maxVars = maxOccurenceVar p
       if (0 < min || 0 > max || currentVars >= maxVars) 
           then return Nothing
           else do (n, w) <- randomIdentifier p varParameter
                   return $ Just $ (w, (getVariableTermType n, 0))

-- | Funktion, die abhngig von der Parameterisierung eine zufllige Kontextvariable
--   liefert. Dabei wird die aktuelle Anzahl bereits generierter und das
--   diesbezgliche Limit von Kontextvariablen bercksichtigt.
feedCVariable :: TermGeneratorParameter -> FeedFunction
feedCVariable p (min, max) = 
    do currentCVars <- gets occurenceCVar
       let maxCVars = maxOccurenceCVar p
       if (1 < min || 1 > max || currentCVars >= maxCVars)
           then return Nothing
           else do (n, w) <- randomIdentifier p cvarParameter
                   return $ Just $ (w, (getContextVariableTermType n, 1))

-- | Funktion, die ein Loch liefert.
feedHole :: TermGeneratorParameter -> FeedFunction
feedHole p (min, max) = 
    if (0 < min || 0 > max)
        then return Nothing
        else do let ttp = holeParameter p
                    w   = termTypeWeight ttp
                return $ Just $ (w, (TermHole, 0))

-- | Funktion, die abhngig von der Parameterisierung einen Funktionssubterm
--   liefert. Dabei wird die geforderte Stelligkeit bercksichtigt.
feedFunction :: TermGeneratorParameter  -- ^ externer Parameter
    -> FunctionTable                    -- ^ Funktionstabelle fr Stelligkeiten der Funktionen
    -> FeedFunction
feedFunction p ft (min, max) = 
    do (n, w) <- randomIdentifier p funcParameter
       let ar = getArity ft n
       if (ar < min || ar > max)
            then feedFunction p ft (min, max)
            else return $ Just $ (w, (getFunctionTermType n, ar))

-- | Whlt eine der bergebenen TermType-Actions und fhrt diese aus.
--   Gewichte werden bercksichtigt
feedTerm :: (Int, Int)             -- ^ geforderte Stelligkeit (min, max)
    -> [TermFeed -> FeedFunction]  -- ^ Funktionen gewnschter Termtypen
    -> GeneratorState RTermType    -- ^ gewhlter Termtyp
feedTerm r ffs =
    do tf       <- gets tsFeed
       rs_maybe <- sequence $ map (\ff -> ff tf r) ffs
       -- bercksichtige nur erfolgreiche Wahlen. 
       let rs   =  catMaybes rs_maybe
       -- Whle anhand des Termgewichtes einen Wert zufllig aus.
       tt       <- lift $ pickFromWeightedIO rs
       return tt

-- | Hilfsfunktion, die zufllig einen Term liefert. Die Funktion
--   berechnet einen zuflligen Termidentifikator (eine Nummer) anhand
--   der Anzahl maximal gewnschter Terme (siehe TermGeneratorParameter).
randomIdentifier :: TermGeneratorParameter  -- ^ externe Parameterisierung 
    -> (TermGeneratorParameter -> TermTypeParameter) -- ^ Funktion, die aus einem TermGeneratorParameter
                                                     --   den TermTypeParameter des gewnschten TermTypes holt.
                                                     --   notwendig fr Gewicht & max. Anzahl TermTypes
    -> GeneratorState (Int, Int)                     -- ^ (gewhlter Termidentifikator, Gewicht des Terms)
randomIdentifier p f = 
    do let tt =  f p
           n  =  termTypeCount tt
           w  =  termTypeWeight tt
       rf     <- getParameter feedRange
       -- Whle zufllig einen Termidentifikator aus der maximalen Menge aus.
       result <- lift $ rf (0, n-1)
       return (result, w)

-- | Liefert anhand eines Termidentifikators einen Funktions-TermType.
getFunctionTermType :: Int  -- ^ Termidentifikator
    -> TermType
getFunctionTermType i = TermFunction ("f" ++ show i)

-- | Liefert anhand eines Termidentifikators einen Variablen-TermType.
getVariableTermType :: Int  -- ^ Termidentifikator
    -> TermType
getVariableTermType i = TermVariable ("v" ++ show i)

-- | Liefert anhand eines Termidentifikators einen KontextVariablen-TermType.
getContextVariableTermType :: Int   -- ^ Termidentifikator
    -> TermType
getContextVariableTermType i = TermContextVariable ("X" ++ show i)

---------------------------------
-- Funktionstabelle
---------------------------------

-- | Funktionstabelle. Diese Funktionstabelle muss vorab berechnet werden,
--   da die Funktionen einheitlich eine Stelligkeit besitzen mssen.
--   Array fr alle Funktionen, Mapping auf Arity
newtype FunctionTable = FunctionTable (Array Int Int)     
    deriving Show

-- | Liefert Stelligkeit einer Funktion.
getArity :: 
    FunctionTable      -- ^ Funktionstabelle mit Stelligkeiten
    -> Int             -- ^ Nummer der Funktion
    -> Int             -- ^ Stelligkeit der angefragten Funktion
getArity (FunctionTable a) i = (a ! i)

-- | Erzeugt eine zufllige Funktionstabelle anhand der bergebenen Parameterisierung
--   Es wird _immer_ mindestens eine Funktion mit Stelligkeit 0 und eine Funktion
--   mit Stelligkeit 1 generiert!
createFunctionTable :: TermGeneratorParameter -> IO FunctionTable
createFunctionTable p@TermGeneratorParameter {
      funcParameter = fp,
      maxArity = ar,
      randomRangeF = rf
      } = 
     do let n            = termTypeCount fp
            randomNumber =  if (n > 2) then n-2 else 0
        randomArity      <- sequence $ take randomNumber $ repeat $ rf (0, ar)
        let arity        =  [0,1] ++ randomArity -- mind ein 0, ein 1-arity Functionterm
        return $ FunctionTable $ array (0, n-1) (zip [0..n-1] arity)

----------------------------------------------------------------------
-- Anzahl finden

-- | Zhlt die Hufigkeiten aller Variablen in dem gegebenen Term
numberVariables :: (Term t) =>
    t               -- ^ zu untersuchender Term
    -> Int          -- ^ Anzahl theoretisch vorhandener Variablen im Term
    -> [(Int, Int)] -- ^ Liste aus (Termidentifikator, Hufigkeit)
numberVariables t n = filter (\(i,n) -> n > 0) $ countTerms t n isVariable

-- | Zhlt die Hufigkeiten aller Kontextvariablen in dem gegebenen Term
numberCVariables :: (Term t) =>
    t               -- ^ zu untersuchender Term
    -> Int          -- ^ Anzahl theoretisch vorhandener Kontextvariablen im Term
    -> [(Int, Int)] -- ^ Liste aus (Termidentifikator, Hufigkeiten
numberCVariables t n = filter (\(i,n) -> n > 0) $ countTerms t n isContextVariable

-- | Hilfsfunktion. Zhlt die Anzahl von Hufigkeiten bestimmter Subtermtypen
--   in einem Term. Die zu bestimmenden Subterme werden mit Hilfe einer bergebenen
--   Prdikatsfunktion ausgewhlt.
--   Die Funktion geht davon aus, dass die Termnamen "[a-zA-Z][0-9]+" aufgebaut sind.
--   Der Termidentifikator wird dann aus der Nummer im Termnamen gewonnen.
countTerms :: (Term t) =>
    t               -- ^ zu untersuchender Term
    -> Int          -- ^ Anzahl theoretisch vorhandener Subterme des betrachteten Typs
    -> (t -> Bool)  -- ^ liefert true, falls der aktuelle Subterm vom betrachteten Typ ist
    -> [(Int, Int)] -- ^ Liste aus (Termidentifikator, Hufigkeit)
countTerms t n test =
    let nameList = map getName $ filter test $ toList t
        numberList = (map (read . tail) nameList) :: [Int]

        occurences :: TreeFiniteMap Int Int
        occurences = occurenceMap numberList

        findOccurence :: TreeFiniteMap Int Int -> Int -> (Int, Int)
        findOccurence m i = 
            case (lookupFM i m) of
                Nothing  -> (i, 0)
                (Just n) -> (i, n)
    in  map (findOccurence occurences) [0..(n-1)]

-- | Berechnet anhand einer Liste eine Map von Element -> Anzahl der
--   Vorkommen des Elements in der Liste
occurenceMap :: (Ord a) => [a] -> TreeFiniteMap a Int
occurenceMap l = foldl (\m j -> updateFM j 1 (\i -> i+1) m) emptyFM l
    
-------------------------------------------------------------------------------
-- Hilfsfunktionen
-------------------------------------------------------------------------------

-- | Typ, der einen TermType mit seiner Stelligkeit vereint
type RTermType = (TermType, Int)    -- ^ (TermType, Stelligkeit)

-- | Funktion, die aus einem Termtyp und weiteren TermNodes (dessen Kindern)
--   einen neuen TermNode erzeugt. Mit Hilfe dieser Funktion wird der Term-
--   baum aufgebaut.
cRNode :: RTermType                      -- ^ TermType des Vaters
    -> [GeneratorState ExtendedTermNode] -- ^ Kinder-TermNodes
    -> GeneratorState ExtendedTermNode   -- ^ der erzeuget Vater-TermNode
cRNode f gchs = do chs <- sequence gchs
                   return $ cTermNode (fst f) chs

-- | Hilfsfunktion. Gibt einen bestimmten Parameter zurck. Beispielsweise
-- gibt @getParameter numVar@ die max. Anzahl Variablen zurck.
getParameter :: (TermFeed -> a) 
    -> GeneratorState a
getParameter f = do feed <- gets tsFeed
                    return (f feed)

-- | Debuggingausgabe auf Konsole
debug :: String -> GeneratorState ()
debug msg = lift $ putStr (msg ++ "\n")


