-- |Definiert den Parser fr Kontextmatchingprobleme. Dies ist die EBNF fr
-- Kontextterme:
--
-- UnifikationsProblem ::= "[" TermEquationList "]"
--
-- TermEquationList ::= TermEquation | TermEquation "," TermEquationList
-- TermEquation ::= Term "=?" Term
--
-- Term ::= Hole | Variable | ContextVariable | Function
-- TermList ::= Term "," TermList | Term
--
-- Hole ::= "_"
-- Variable ::= Identifier
-- ContextVariable ::= Identifier "[" Term "]"
-- Function ::= Identifier "(" TermList ")"
-- 
-- Identifier ::= [a..zA..Z0..9]+
--
-- Wir verwenden PARSEC, einen Parserkombinatorlibrary.
-- Innerhalb der Terme wird whitespace ignoriert. Haskell-Kommentare sind
-- erlaubt.
--
-- Referenzen: http://www.cs.uu.nl/people/daan/parsec.html
module Gersdorf.ContextMatching.Base.TermParser (
    unificationProblem,
    parseUPError,
    module Gersdorf.ContextMatching.Base.Terms
) where

import Text.ParserCombinators.Parsec
import qualified Text.ParserCombinators.Parsec.Token as P
import Text.ParserCombinators.Parsec.Language

import Gersdorf.ContextMatching.Base.Terms
import Gersdorf.ContextMatching.Base.SimpleTermEquation
import Gersdorf.ContextMatching.Base.SimpleUnificationProblem

-- Wir machen SimpleUnificationProblems zu Instanzen von Read.
instance Read ContextUP where
    readsPrec _ s = parseIt unificationProblem s
instance Read ContextTerm where
    readsPrec _ s = parseIt term s
instance Read ContextTermEq where
    readsPrec _ s = parseIt termEquation s

parseIt :: Parser a -> String -> [(a, String)]
parseIt p s = case (parse p "" s) of
                Left err -> []
                Right x -> [(x, "")]

parseUPError :: String -> Either ParseError ContextUP
parseUPError s = parse unificationProblem "" s

-- Definition des Lexers. Er wird benutzt, um whitespace zu ignorieren.
lexer :: P.TokenParser ()
lexer = P.makeTokenParser haskellDef 

-- Wird definieren Lexikalisch Parser.
whiteSpace = P.whiteSpace lexer
lidentifier = P.identifier lexer
lexeme = P.lexeme lexer

-- Dies ist die Wurzel aller Unifikationsprobleme.
unificationProblem :: Parser ContextUP 
unificationProblem = do whiteSpace
                        lexeme (char '[')
                        l <- termEquationList
                        lexeme (char ']')
                        return (cProblem l)

termEquationList :: Parser [ContextTermEq]
termEquationList = do ll <- sepBy1 termEquation separator
                      return ll

termEquation :: Parser ContextTermEq
termEquation = do t1 <- term
                  lexeme (string "=?")
                  t2 <- term
                  return (cEquation t1 t2)

term :: Parser ContextTerm
term = do tn <- innerTerm
          return (cTerm tn)

innerTerm :: Parser ContextTermNode
innerTerm = do ll <- (hole 
                 <|> try contextVariable 
                 <|> try function
                 <|> variable
                 )
               return ll

termList :: Parser [ContextTermNode]
termList = do terms <- sepBy innerTerm separator
              return terms

hole :: Parser ContextTermNode
hole = do lexeme (char '_')
          return cHole

variable :: Parser ContextTermNode
variable = do id <- identifier
              return (cVar id)

contextVariable :: Parser ContextTermNode
contextVariable = do id <- identifier
                     lexeme (char '[')
                     subterm <- innerTerm
                     lexeme (char ']')
                     return (cContextVar id subterm)

function :: Parser ContextTermNode
function = do id <- identifier
              lexeme (char '(')
              arguments <- termList
              lexeme (char ')')
              return (cFunction id arguments)

separator = skipMany1 (space <|> char ',')

identifierList :: Parser [String]
identifierList = do ids <-sepBy1 identifier sep
                    return ids
  where sep = skipMany1 (space <|> char ',')

identifier :: Parser String
identifier = do id <- lidentifier
                return id

run :: Show a => Parser a -> String -> IO ()
run p input = case (parse p "" input) of
                Left err -> do { putStr "parseErr at"
                               ; print err
                               }
                Right x -> print x

testTermParse :: String -> IO ()
testTermParse input = run unificationProblem input
