module Gersdorf.ContextMatching.General.GeneralCM (
    solveGeneralCM,
    solveGeneralCMStatistic,
    solveOneStepCM,
    solveBacktrackGCM,
    solveBacktrackGCMVerbose,
    ComplexNode(..),
    BacktrackingStatistic(..),
    RuleStep(..)
) where

import Gersdorf.ContextMatching.Base.UnificationProblem
import Gersdorf.ContextMatching.Base.Substitution
import Gersdorf.Algorithms.Backtracking
import Gersdorf.Data.BacktrackingChoice

import Gersdorf.ContextMatching.General.EnvEquationSet
import Gersdorf.ContextMatching.General.Heuristic

import Gersdorf.ContextMatching.General.Rules.Identity

import Gersdorf.ContextMatching.Base.Utilities

import Debug.Trace

solveGeneralCM :: (EnvPosEqSet es e t tp, UnificationProblem up e t) => 
    Strategy es e t tp
    -> up e t 
    -> [Substitution t]
solveGeneralCM str up =
    let equations = toEquationList up
        eset      = cSet equations

        solutions  = solveBacktrackGCMVerbose str (generate . ccCurrentNode) (\r _ -> [r]) eset
        flatSubsts = concat solutions

    in  flatSubsts

solveBacktrackGCM :: (EnvPosEqSet es e t tp) =>
	Strategy es e t tp
	-> es e t
	-> [es e t]
solveBacktrackGCM str eset = 
    solveBacktrackGCMVerbose str ccCurrentNode (:) eset

solveGeneralCMStatistic :: (EnvPosEqSet es e t tp) =>
    Strategy es e t tp
    -> es e t
    -> BacktrackingStatistic (Rule es e t tp)
solveGeneralCMStatistic str eset =
    let firstNode = StatisticNode { snCurrentNode = eset, snLastRule = identityRule }
    in  solveBacktrackStatistic (calculateNextStepStatistic str) 
                                (isEmpty . snCurrentNode)
                                (\statisticNode -> snLastRule statisticNode)
                                firstNode

calculateNextStepStatistic :: (EnvPosEqSet es e t tp) =>
	Strategy es e t tp
	-> StatisticNode es e t tp
	-> BacktrackingChoice (StatisticNode es e t tp)
calculateNextStepStatistic str (StatisticNode { snCurrentNode = eset }) =
    let (resultRule, resultBT) = calculateNextStep str eset
        
        result =  fmap (\newEset -> StatisticNode { snCurrentNode = newEset, snLastRule = resultRule })
                       resultBT
    in  result

data  (EnvPosEqSet es e t tp) => StatisticNode es e t tp =
    StatisticNode {
        snCurrentNode :: es e t,
        snLastRule    :: Rule es e t tp
    }

data (EnvPosEqSet es e t tp) => RuleStep es e t tp =
    RuleStep {
        rsRule     :: Rule es e t tp,
        rsSnapshot :: es e t
    }
instance (EnvPosEqSet es e t tp) => Show (RuleStep es e t tp) where
    showsPrec _ rs r = showString ("[" ++ show (rsRule rs) ++ "] ")
                                  (shows (rsSnapshot rs) r)

cRuleStep :: (EnvPosEqSet es e t tp) =>
    Rule es e t tp
    -> es e t
    -> RuleStep es e t tp
cRuleStep rule eset = RuleStep { rsRule = rule, rsSnapshot = eset }

data  (EnvPosEqSet es e t tp) => ComplexNode es e t tp =
    ComplexNode {
        ccCurrentNode :: es e t,
        ccSteps :: !([RuleStep es e t tp]),
        ccTakenSteps :: !Int  -- strict to avoid unnecessary heap growth due to addComplexNode!
    }
instance (EnvPosEqSet es e t tp) => Show (ComplexNode es e t tp) where
    showsPrec _ cn r = 
        showString ("\nComplexNode. Taken: " ++ show (ccTakenSteps cn) ++ ". \nCurrent: ")
                   (shows (ccCurrentNode cn)
                          (showPredList "\nSteps:\n" (ccSteps cn) "\n" "" r))

cComplexNode :: (EnvPosEqSet es e t tp) =>
    [RuleStep es e t tp]
    -> Int
    -> es e t
    -> ComplexNode es e t tp
cComplexNode steps taken node = 
    ComplexNode { 
        ccCurrentNode = node, 
        ccSteps = steps, 
        ccTakenSteps = taken 
    }

reverseComplexNode :: (EnvPosEqSet es e t tp) =>
    ComplexNode es e t tp
    -> ComplexNode es e t tp
reverseComplexNode cn = cComplexNode (reverse $ ccSteps cn) 
                                     (ccTakenSteps cn)
                                     (ccCurrentNode cn)

addComplexNode :: (EnvPosEqSet es e t tp) =>
     ComplexNode es e t tp
     -> Rule es e t tp
     -> (RuleStep es e t tp -> [RuleStep es e t tp] -> [RuleStep es e t tp])
     -> es e t
     -> ComplexNode es e t tp
addComplexNode (ComplexNode { ccSteps = steps, ccTakenSteps = taken }) rule nodeadder node =
    let rulestep = cRuleStep rule node
    in ComplexNode {
          ccCurrentNode = node,
          ccSteps = nodeadder rulestep steps, 
          ccTakenSteps = taken + 1
       }


solveBacktrackGCMVerbose :: (EnvPosEqSet es e t tp, Show a) =>
	Strategy es e t tp
    -> (ComplexNode es e t tp -> a)
    -> (RuleStep es e t tp -> [RuleStep es e t tp] -> [RuleStep es e t tp])
	-> es e t
	-> [a]
solveBacktrackGCMVerbose str f nodeadder eset = 
    let firstNode = cComplexNode [cRuleStep identityRule eset] 0 eset
    in  solveBacktrackSolution (calculateNextStepComplex str nodeadder) 
                               (isEmpty . ccCurrentNode)
                               (f . reverseComplexNode)
                               firstNode
        -- head ist hier erlaubt, da die Liste nie leer ist, da zu Anfang
        -- eine identityRule eingefgt wird (s. firstNode)

instance (EnvPosEqSet es e t tp) => Eq (ComplexNode es e t tp) where
    (ComplexNode { ccCurrentNode = a })
     == (ComplexNode { ccCurrentNode = b }) = a == b

-- removes bot choices
removeBotBacktrackingChoice :: (EnvPosEqSet es e t tp) =>
    BacktrackingChoice (es e t)
    -> BacktrackingChoice (es e t)
removeBotBacktrackingChoice bt =
    fmapRemove (\eset -> if (isEqBot eset) then Nothing else Just eset) bt

calculateNextStepComplex :: (EnvPosEqSet es e t tp) =>
	Strategy es e t tp
    -> (RuleStep es e t tp -> [RuleStep es e t tp] -> [RuleStep es e t tp])
	-> ComplexNode es e t tp
	-> BacktrackingChoice (ComplexNode es e t tp)
calculateNextStepComplex str nodeadder cn@(ComplexNode { ccCurrentNode = eset }) =
    let (resultRule, resultBT) = calculateNextStep str eset
        
        result   =  fmap (addComplexNode cn resultRule nodeadder) resultBT
    in result
--	in trace ("=> Rule: " ++ show resultRule ++ " //// " ++ show resultBT ++ " <=") result

calculateNextStep :: (EnvPosEqSet es e t tp) =>
	Strategy es e t tp
	-> es e t 
	-> (Rule es e t tp, BacktrackingChoice (es e t))
calculateNextStep str eset =
    let choice     = runStrategy str eset
        resultRule = chosenRule choice
        resultBT   = removeBotBacktrackingChoice (chosenResult choice)
        result   =  if (noChoice resultBT) 
        				then cBacktrackingNo 
                        else resultBT
        
    in (resultRule, result)
    
solveOneStepCM :: (UnificationProblem up e t, EnvPosEqSet es e t tp) => 
	Strategy es e t	tp
	-> up e t 
	-> HeuristicChoice es e t tp
solveOneStepCM str up =
   let equations = toEquationList up
       eset      = (cSet equations) 
   in  runStrategy str eset


