module Gersdorf.ContextMatching.General.GeneralFacade (
    solveGCM,
    solveGCMVerbose,
    solveGCMStatistic, solveGCMAggregatedStatistic,
    testHeuristic,
    applyStrategyRules,
    applyRuleUP,

    applyIO,
    applyIOn,

    getWrongSolutions,
    allSolutionsOkay,
    regressionTest,

    statisticTestDiff,
    StatisticInformationDiff(..),
    AggregatedStatisticInformation(..),
    StatisticInformation(..)
) where

import Control.Monad

import Gersdorf.Data.List
import Gersdorf.Data.FiniteMap

import Gersdorf.ContextMatching.General.GeneralCM
import Gersdorf.ContextMatching.General.Rules.Rules
import Gersdorf.ContextMatching.General.Heuristic

import Gersdorf.ContextMatching.Linear.LinearityTest (extractTerms)

import Gersdorf.ContextMatching.Base.Terms
import Gersdorf.ContextMatching.Base.HeapTerm
import Gersdorf.ContextMatching.Base.SimpleTermEquation
import Gersdorf.ContextMatching.Base.SimpleUnificationProblem
import Gersdorf.ContextMatching.Base.ExtendedTerm

import Gersdorf.ContextMatching.Base.Utilities

---------------------------------------------------------------------------------------
-- Lsungsfunktionen
---------------------------------------------------------------------------------------

-- | Lst das bergebene Problem mit der gegebenen Strategie und gibt
--   ausfhrliche Informationen ber die Lsungen (incl. Lsungsschritte) zurck.
solveGCMVerbose :: (EnvPosEqSet SimpleEnvEquationSet eq t tp, UnificationProblem up eq t) =>
	Strategy SimpleEnvEquationSet eq t tp -- ^ anzuwendende Strategie
	-> up eq t	                          -- ^ Unifikationsproblem
	-> [ComplexNode SimpleEnvEquationSet eq t tp]
solveGCMVerbose str up = 
    let equations = toEquationList up
        eset      = cSet equations
    in  solveBacktrackGCMVerbose str id (:) eset

-- | Lst das bergebene Problem mit der gegebenen Strategie und gibt die
--   Lsungssubstitutionen zurck.
solveGCM :: (EnvPosEqSet SimpleEnvEquationSet eq t tp, UnificationProblem up eq t) =>
    Strategy SimpleEnvEquationSet eq t tp  -- ^ anzuwendende Strategie
	-> up eq t	                           -- ^ Unifikationsproblem
	-> [Substitution t]
solveGCM str up = solveGeneralCM str up

solveGCMStatistic :: (EnvPosEqSet SimpleEnvEquationSet eq t tp, UnificationProblem up eq t) =>
    Strategy SimpleEnvEquationSet eq t tp  -- ^ anzuwendende Strategie
    -> up eq t	                           -- ^ Unifikationsproblem
    -> StatisticInformation SimpleEnvEquationSet eq t tp
solveGCMStatistic str up = 
    let equations  = toEquationList up
        eset       = cSet equations
        statistics = solveGeneralCMStatistic str eset
        si         = calcStatisticInformation up statistics
    in  si

solveGCMAggregatedStatistic :: (UnificationProblem up eq t, Substitutable (up eq t),
                      (EnvPosEqSet SimpleEnvEquationSet eq t tp)) =>
    Strategy SimpleEnvEquationSet eq t tp  -- ^ Anzuwendende Strategie
    -> [up eq t]                           -- ^ Menge von zu lsenden Problemen
    -> AggregatedStatisticInformation SimpleEnvEquationSet eq t tp
solveGCMAggregatedStatistic str ups =
    snd $ statisticTestComplete str ups

-- | Fhrt die angegebene Strategie und die in ihr enthaltene Heuristik 
--   einen Schritt aus und und gibt das Ergebnis zurck.
testHeuristic :: (EnvPosEqSet SimpleEnvEquationSet e t tp, UnificationProblem up e t) =>
    Strategy SimpleEnvEquationSet e t tp   -- ^ anzuwendende Strategy
	-> up e t                              -- ^ Unifikationsproblem
	-> HeuristicChoice SimpleEnvEquationSet e t tp
testHeuristic str up = solveOneStepCM str up

-- | Wendet auf das bergebene Problem alle in der gegebenen Strategie
--   vorhandenen Regeln an und gibt jeweils das Ergebnis zurck.
applyStrategyRules :: (EnvPosEqSet SimpleEnvEquationSet eq t tp, UnificationProblem up eq t) =>
	Strategy SimpleEnvEquationSet eq t tp  -- ^ anzuwendende Strategie
	-> up eq t                             -- ^ Unifikationsproblem
	-> [RuleApplicationResult SimpleEnvEquationSet eq t tp]
applyStrategyRules (Strategy { strategyRules = rules }) up =
	let equations = toEquationList up
            eset      = cSet equations
        in  map (applyRule eset) rules

applyRuleUP :: (EnvPosEqSet SimpleEnvEquationSet eq t tp, UnificationProblem up eq t) =>
    Rule SimpleEnvEquationSet eq t tp
	-> up eq t                             -- ^ Unifikationsproblem
	-> RuleApplicationResult SimpleEnvEquationSet eq t tp
applyRuleUP r up =
    let equations = toEquationList up
        eset      = cSet equations
    in  applyRule eset r


-- | Hilfsfunktion, um eine Strategiefunktion anzuwenden auf ein Unifikationsproblem,
--   das aus der IO-Monade kommt.
--   Beispiel: applyIO solveGCM baseStrategy (generateUP varRTP)
applyIO :: (Show a, UnificationProblem up eq t, EnvPosEqSet SimpleEnvEquationSet eq t tp) => 
   (Strategy SimpleEnvEquationSet eq t tp -> up eq t -> a)  -- ^ Eigentlich anzuwendende pure Funktion
   -> Strategy SimpleEnvEquationSet eq t tp                 -- ^ anzuwendende Strategie
   -> (IO (up eq t))                                        -- ^ Action, die ein Unifikationsproblem liefert
   -> IO a
applyIO solF str ioUP =
   do up      <- ioUP
      let sol =  solF str up
      putStr ("Problem: " ++ show up ++ "\n")
      putStr ("Solution: " ++ show sol ++ "\n")
      return sol

-- | Hilfsfunktion, um eine Strategiefunktion anzuwenden, die auf n Unifikationsproblemen
--   operiert, und einer Funktion, die innerhalb der IO-Monade n Unifikationsprobleme
--   generiert.
--   Beispiel: applyIOn statisticTestComplete baseStrategy (generateUPn varRTP) 20
applyIOn :: (Show a, UnificationProblem up eq t, EnvPosEqSet SimpleEnvEquationSet eq t tp) => 
   (Strategy SimpleEnvEquationSet eq t tp -> [up eq t] -> a) -- ^ eigentlich anzuwendende pure Funktion
   -> Strategy SimpleEnvEquationSet eq t tp                  -- ^ anzuwendende Strategie
   -> (Int -> IO [up eq t])                                  -- ^ Funktion, die in der IO-Monade n
                                                             --   Unifikationsprobleme berechnet
   -> Int                                                    -- ^ Anzahl der Probleme, die verwendet
                                                             --   werden sollen
   -> IO a
applyIOn solF str ioUP n =
   do ups     <- ioUP n
      let sol =  solF str ups
      putStr ("Solution: " ++ show sol ++ "\n")    
      return sol
      
------------------------------------------------------------------------------
-- Testfunktionen 
------------------------------------------------------------------------------

-- | Lst das gegebene Problem mit der gebenenen Strategie und prft,
--   ob die errechneten Lsungen korrekt sind. Dies ist kein Test auf 
--   Vollstndigkeit der Lsung sondern nur auf Korrektheit der Lsung!
--
--   Falsche Lsungen werden in der Rckgabeliste zurckgegeben.
getWrongSolutions :: (UnificationProblem up eq t, Substitutable (up eq t),
                 (EnvPosEqSet SimpleEnvEquationSet eq t tp)) =>
	Strategy SimpleEnvEquationSet eq t tp    -- ^ Anzuwendende Strategie
	-> up eq t	                             -- ^ Unifikationsprobleme
	-> [Substitution t]                      -- ^ Lsungen, die falsch sind
getWrongSolutions str up = 
    let solutions   = solveGCM str up
        isSubstOkay = \subst -> isCorrectUnderSubst up subst
    in  filter (not . isSubstOkay) solutions

-- | Lst das gegebene Problem mit der gebenenen Strategie und prft,
--   ob die errechneten Lsungen korrekt sind. Dies ist kein Test auf 
--   Vollstndigkeit der Lsung sondern nur auf Korrektheit der Lsung!
allSolutionsOkay :: (UnificationProblem up eq t, Substitutable (up eq t),
                 (EnvPosEqSet SimpleEnvEquationSet eq t tp)) =>
	Strategy SimpleEnvEquationSet eq t tp    -- ^ Anzuwendende Strategie
	-> up eq t	                             -- ^ Unifikationsproblem
	-> Bool                                  -- ^ True, falls alle Lsungen korrekt sind.
allSolutionsOkay str up = null $ getWrongSolutions str up

-- | Lst alle in der Eingabe vorhandenen Probleme und prft die errechneten
--   Lsungen auf Korrektheit. Probleme, die nicht korrekt gelst wurden,
--   werden zurckgegeben.
regressionTest :: (UnificationProblem up eq t, Substitutable (up eq t),
                 (EnvPosEqSet SimpleEnvEquationSet eq t tp)) =>
    Strategy SimpleEnvEquationSet eq t tp   -- ^ Anzuwendende Strategie
    -> [up eq t]                            -- ^ Unifikationsproblemmenge
    -> [up eq t]                            -- ^ inkorrekt gelste Probleme
                                            --   (sollte leer sein!)
regressionTest str ups
    = filter (not . allSolutionsOkay str) ups

--------------------------------------------------------------------------
-- Statistiktests
--------------------------------------------------------------------------

-- | Vergleichende Statistikinformationen ber die Lsung einer Menge
--   von Problemen mittels zweier Strategien
data (EnvPosEqSet es eq t tp) => StatisticInformationDiff es eq t tp = 
    StatisticInformationDiff {
        -- Ergebnis der Strategieanwendung von Strategie 1
        sidStats1    :: (AggregatedStatisticInformation es eq t tp),
        -- Ergebnis der Strategieanwendung von Strategie 2
        sidStats2    :: (AggregatedStatisticInformation es eq t tp),
        -- Differenz in der Anzahl der Lsungen (sollte stets 0 sein!!)
        sidSolutionDiff :: Int,
        -- Anzahl der Steps, um die Strategie 1 maximal besser war als Strategie 2
        sidMaxBetter    :: Int,
        -- Anzahl der Probleme, in denen Strategie 1 besser war als Strategie 2
        sidNumBetter    :: Int,
        -- durchschnittliche Anzahl der Steps in den Problemen, in denen Strategie 1 besser als Strategie 2 war
        sidAvgBetter    :: Double,
        -- Anzahl der Steps, um die Strategie 1 maximal schlechter war als Strategie 2
        sidMaxWorse     :: Int,
        -- Anzahl der Probleme, in denen Strategie 1 schlechter war als Strategie 2
        sidNumWorse     :: Int,
        -- durchschnittliche Anzahl der Steps in den Problemen, in denen Strategie 1 schlechter war als Strategie 2
        sidAvgWorse     :: Double
    } deriving Show

{-
instance (EnvPosEqSet es eq t tp) => Show (StatisticInformationDiff es eq t tp) where
    showsPrec _ sid r = 
        showRecordHelper "Statistic Information Difference"
            [("Stats1", show . sidStats1),
             ("Stats2", show . sidStats2),
             ("maxBetter", show . sidMaxBetter),
             ("numBetter", show . sidNumBetter),
             ("avgBetter", show . sidAvgBetter),
             ("maxWorse", show . sidMaxWorse),
             ("numWorse", show . sidNumWorse),
             ("avgWorse", show . sidAvgWorse)]
            sid
            r
-}

-- | Statistikinfomationen ber die Lsung einer Menge von Problemen mit einer
--   einzigen Strategie
data (EnvPosEqSet es eq t tp) => AggregatedStatisticInformation es eq t tp = 
    AggregatedStatisticInformation {
        -- Insgesamt ausgefhrte Regelschritte (ber alle Probleme)
        asiTotalSteps :: Int,
        -- Insgesamt errechnete Lsungen (ber alle Probleme)
        asiTotalSolutions :: Int,
        -- Durchschnittliche Problemgre
        asiAvgSize :: Double,
        -- Durchschnittliche Regelschrittanzahl pro Problem
        asiAvgSteps :: Double,
        -- Statistische Auswertung ber die Anzahl der Anwendung jeder Regel
        -- ber alle Probleme
        asiRuleSteps :: [(Rule es eq t tp, Int)],
        
        asiAvgFunctionCount :: Double,
        asiAvgDifferentFunctionCount :: Double,
        asiAvgHoleCount :: Double,
        asiAvgVariableCount :: Double,
        asiAvgDifferentVariableCount :: Double,
        asiAvgCVariableCount :: Double,
        asiAvgDifferentCVariableCount :: Double
    } deriving Show
    
{-
instance (EnvPosEqSet es eq t tp) => Show (AggregatedStatisticInformation es eq t tp) where
    showsPrec _ asi r = 
        showRecordHelper "Aggregated Statistic Information"
            [("total Steps", show . asiTotalSteps),
             ("avg Steps/Problem", show . asiAvgSteps),
             ("Rule overview", show . asiRuleSteps)]
            asi
            r
-}

-- | Statistikinformation ber die Anwendung einer Strategie auf ein
--   Unifikationsproblem
data (EnvPosEqSet es eq t tp) => StatisticInformation es eq t tp = 
    StatisticInformation {
        -- Anzahl insgesamt ausgefhrter Schritte
        siTotalSteps :: Int,
        -- Anzahl insgesamt errechneter Lsungen
        siTotalSolutions :: Int,
        -- Statistische Auswertung ber die Anzahl der Anwendung jeder Regel
        -- ber alle Probleme
        siRuleSteps :: [(Rule es eq t tp, Int)],
 
        -- Statistiken ber das Unifikationsproblem
        siFunctionCount :: Int,
        siDifferentFunctionCount :: Int,
        siHoleCount :: Int,
        siVariableCount :: Int,
        siDifferentVariableCount :: Int,
        siCVariableCount :: Int,
        siDifferentCVariableCount :: Int
    } deriving Show

{-
instance (UnificationProblem up eq t) => Show (StatisticInformation up eq t) where
    showsPrec _ si r = 
        showRecordHelper "Statistic Information"
            [("Problem", show . siProblem),
             ("total Steps", show . siTotalSteps),
             ("total solutions", show . siTotalSolutions)]
            si
            r
-}

-- | Berechnet Vergleichende Statistikinformationen ber die Lsung einer Menge
--   von Problemen mittels zweier Strategien
statisticTestDiff :: (UnificationProblem up eq t, Substitutable (up eq t),
                 (EnvPosEqSet SimpleEnvEquationSet eq t tp)) =>
    Strategy SimpleEnvEquationSet eq t tp     -- ^ Strategie 1
    -> Strategy SimpleEnvEquationSet eq t tp  -- ^ Strategie 2
    -> [up eq t]                              -- ^ Menge von Problemen
    -> StatisticInformationDiff SimpleEnvEquationSet eq t tp
statisticTestDiff str1 str2 ups =
    let asi1   = statisticTestComplete str1 ups     -- zunchst: berechne globale Statistikinformationen
        asi2   = statisticTestComplete str2 ups
        steps1 = map siTotalSteps $ fst asi1        -- extrahiere die Anzahl der Schritte pro Problem
        steps2 = map siTotalSteps $ fst asi2
        stepsPair = zip steps1 steps2

        -- berechne Differenz der Anzahl der Schritte pro Lsung
        diffs = map (\(s1, s2) -> s1 - s2) stepsPair
        -- Extrahiere bessere/schlechtere Anzahl der Schritte pro Lsung
        betters = map abs $ filter ((>) 0) diffs
        worses  = filter ((<) 0) diffs
    in  StatisticInformationDiff {
            sidStats1    = snd asi1,
            sidStats2    = snd asi2,
            sidSolutionDiff = asiTotalSolutions (snd asi1) - asiTotalSolutions (snd asi2),
            sidMaxBetter = maximum (0:betters),
            sidNumBetter = length betters,
            sidAvgBetter = calcAverage betters,
            sidMaxWorse  = maximum (0:worses),
            sidNumWorse  = length worses,
            sidAvgWorse  = calcAverage worses
        } 


-- | Berechnet Statistikinfomationen ber die Lsung einer Menge von Problemen mit einer
--   einzigen Strategie. Liefert Informationen ber jedes einzelne Problem (erste Komponente
--   des Ergebnispaars) als auch aggregierte Informationen ber alle Probleme (zweite
--   Komponente des Ergebnispaars).
statisticTestComplete :: (UnificationProblem up eq t, Substitutable (up eq t),
                 (EnvPosEqSet SimpleEnvEquationSet eq t tp)) =>
    Strategy SimpleEnvEquationSet eq t tp  -- ^ Anzuwendende Strategie
    -> [up eq t]                           -- ^ Menge von zu lsenden Problemen
    -> ([StatisticInformation SimpleEnvEquationSet eq t tp], AggregatedStatisticInformation SimpleEnvEquationSet eq t tp)
statisticTestComplete str ups =
    let esets      = map (cSet . toEquationList) ups

        statistics = map (solveGeneralCMStatistic str) esets
        flatOccurences = concat $ map (fmToList . btsTransformMap) statistics

        numSteps :: [Int]
        numSteps = map btsSteps statistics

        numSolutions :: [Int]
        numSolutions = map btsNoSolutions statistics

        sizes :: [Int]
        sizes = map ((foldl (+) 0) . map sizeEq . toEquationList ) ups

        occurences = joinEquals flatOccurences (+)

        -- berechne Einzel-Statistikinformationen
        si  = map (\(up, res) -> calcStatisticInformation up res) $ zip ups statistics
        -- berechne aggregierte Statistikinformationen
        asi = AggregatedStatisticInformation {
                asiTotalSteps = foldl (+) 0 numSteps,
                asiTotalSolutions = foldl (+) 0 numSolutions,
                asiAvgSize    = calcAverage sizes,
                asiAvgSteps   = calcAverage numSteps,
                asiRuleSteps  = occurences,
                
                asiAvgFunctionCount = calcAverage $ map siFunctionCount si,
                asiAvgDifferentFunctionCount = calcAverage $ map siDifferentFunctionCount si,
                asiAvgHoleCount = calcAverage $ map siHoleCount si,
                asiAvgVariableCount = calcAverage $ map siVariableCount si,
                asiAvgDifferentVariableCount  = calcAverage $ map siDifferentVariableCount si,
                asiAvgCVariableCount  = calcAverage $ map siCVariableCount si,
                asiAvgDifferentCVariableCount  = calcAverage $ map siDifferentCVariableCount si
              }
    in  (si, asi)

-- | Baut Statistikinformationen auf fr ein bereits gelstes Unifikationsproblem
--   und berechnet aus der gestellten Lsung Statistikinformationen
calcStatisticInformation :: (EnvPosEqSet SimpleEnvEquationSet eq t tp, UnificationProblem up eq t) =>
    up eq t   -- ^ Unifikationsproblem (nachrichtlich)
    -> BacktrackingStatistic (Rule SimpleEnvEquationSet eq t tp)
    -> StatisticInformation SimpleEnvEquationSet eq t tp
calcStatisticInformation up btstatistics =
    let noSolutions = btsNoSolutions btstatistics
        totalSteps  = btsSteps btstatistics

        terms = extractTerms up
        functions = map getName $ filter isFunction terms
        variables = map getName $ filter isVariable terms
        cvariables = map getName $ filter isContextVariable terms
        holeCount = length $ filter isHole terms
        
    in  StatisticInformation {
            siTotalSteps = totalSteps,
            siTotalSolutions = noSolutions,
            siRuleSteps = fmToList (btsTransformMap btstatistics),
            siFunctionCount = length functions,
            siDifferentFunctionCount = length (countOccurences functions),
            siHoleCount = holeCount,
            siVariableCount = length variables,
            siDifferentVariableCount = length (countOccurences variables),
            siCVariableCount = length cvariables,
            siDifferentCVariableCount = length (countOccurences cvariables)
         }


