module Gersdorf.ContextMatching.General.Rule (
    Rule(..),
    RuleApplicationResult(..),
    applyRule,
    RuleFunction(..),
    BacktrackingChoice,
    cBacktrackingOne, 
    cBacktrackingMany, 
    cBacktrackingDet, 
    cSimpleRule, cSimpleRule1, cSimpleRuleDet,
    module Gersdorf.ContextMatching.General.EnvEquationSet,
    PositionTerm
) where

import Gersdorf.ContextMatching.General.EnvEquationSet
import Gersdorf.ContextMatching.Base.Substitution
import Gersdorf.ContextMatching.Base.PositionTerm
import Gersdorf.Data.BacktrackingChoice

-- | Eine Regel besitzt einen Namen sowie eine Funktion, die das Ergebnis
-- der Regelanwendung berechnet.
data (EnvPosEqSet es eq t tp) => Rule es eq t tp = Rule { 
    ruleName :: String,
    ruleF    :: RuleFunction  es eq t tp
}

instance (EnvPosEqSet es eq t tp) => Eq (Rule es eq t tp) where
    r1 == r2 = (ruleName r1) == (ruleName r2)

instance (EnvPosEqSet es eq t tp) => Ord (Rule es eq t tp) where
    r1 < r2  = (ruleName r1) < (ruleName r2)
    r1 <= r2 = (r1 < r2) || (r1 == r2)

instance (EnvPosEqSet es eq t tp) => Show (Rule es eq t tp) where
	showsPrec _ (Rule { ruleName = name }) r = showString name r

-- |Ein RuleApplicationResult ist Eingabe fr eine Heuristik-Funktion.
-- Es modelliert eine mgliche Regelanwendung. Dabei wird die angewandte
-- Regel festgehalten, ebenso wie das Ergebnis der Regelanwendung. Da
-- dieses don't know-nichtdeterministisch sein kann, wird das Ergebnis
-- in einer BacktrackingChoice gekapselt
data (EnvPosEqSet es eq t tp) => RuleApplicationResult es eq t tp =
    RuleApplicationResult {
        appliedRule :: Rule es eq t tp,                  -- ^ Angewendete Regel
        appliedResult :: [BacktrackingChoice (es eq t)]  -- ^ Ergebnis der Regelanwendung.
    } deriving Show

applyRule :: (EnvPosEqSet es eq t tp) =>
	es eq t
	-> Rule es eq t tp
	-> RuleApplicationResult es eq t tp
applyRule eset r = 
	let (RuleFunction rf) = ruleF r
	in  RuleApplicationResult {
    	 	appliedRule   = r,
	    	appliedResult = filter (not . noChoice) (rf eset)
	    }

-- |Eine RuleFunction liefert - falls die Regel anwendbar ist -
-- ausgehend von einer Gleichungsmenge eine Liste der mglichen, durch
-- Anwendung der Regel auf der Eingabe hervorgehenden, neuen Gleichungsmengen.
--
-- Falls die Regel nicht anwendbar ist, so wird eine leere Liste 
-- zurckgeliefert.
--
-- Falls die Anwendung einer Regel don't know-nichtdeterministisch ist,
-- so wird eine Backtrackingchoice mit Kardinalitt > 1 zurckgeliefert.
--
-- Die Liste im Resultat modelliert verschiedene Regelanwendungsergebnisse,
-- die durch Anwendung auf unterschiedliche Gleichungen in der Eingabegleichungsmenge
-- entstehen.
newtype (EnvEquationSet es eq t, PositionTerm t tp) => RuleFunction es eq t tp = 
    RuleFunction (es eq t                  -- Eingabegleichungsmenge
        -> [BacktrackingChoice (es eq t)]  -- Liste aller mglichen transformierten
      )                                    --   Resultate. 

                                       

-- |Eine SimpleRule ist eine Regel, die bei der Betrachtung eines EquationSets
-- jeweils nur eine Gleichung und evtl. die Umgebung der gelsten Variablen betrachtet.
-- Don't know-Nichtdeterminismus wird untersttzt, indem die Analysefunktion @f@ eine
-- Backtrackingchoice mit mehreren mglichen Ergebnissen zurckliefern kann.
cSimpleRule :: (EnvPosEqSet es eq t tp) =>
    String                          -- Name der Regel
    -> (es eq t -> eq t -> [BacktrackingChoice (RuleResult eq t)]) -- Funktion, die auf zwei Termen operiert und ggf. Lsung liefert
    -> Rule es eq t tp                 -- erstellte Regel
cSimpleRule n f = Rule {
    ruleName = n,
    ruleF = RuleFunction simpleRuleWrapper
    }
    where 
    simpleRuleWrapper eset = scanUpdate eset (f eset)

-- | wie @cSimpleRule@. Die Analysefunktion @f@ arbeitet wie dort auf einem EquationSet und eine Gleichung (Term
-- links und rechts), aber liefert auch nur eine Menge von Gleichungen zurck, d. h. es gibt bei der Anwendung
-- der Regel nur eine Mglichkeit (in Form mehrerer resultierender Gleichungen).
-- Eine solche Regel untersttzt keinen don't know-Nichtedeterminismus.
cSimpleRule1 :: (EnvPosEqSet es eq t tp) =>
    String                          -- Name der Regel
    -> (es eq t -> eq t -> RuleResult eq t) -- Funktion, die auf zwei Termen operiert und ggf. Lsung liefert
    -> Rule es eq t tp                 -- erstellte Regel
cSimpleRule1 n f = let f' eset eq = [f eset eq]
                   in  cSimpleRuleDet n f'

-- | wie @cSimpleRule@. Die Analysefunktion @f@ arbeitet wie dort auf einem EquationSet und eine Gleichung (Term
-- links und rechts), aber liefert auch nur eine Liste von Mengen von Gleichungen zurck, 
-- d. h. es gibt bei der Anwendung der Regel mehrere deterministische Mglichkeiten 
-- (in Form mehrerer resultierender Gleichungen).
-- Eine solche Regel untersttzt keinen don't know-Nichtedeterminismus.
cSimpleRuleDet :: (EnvPosEqSet es eq t tp) =>
    String                          -- Name der Regel
    -> (es eq t -> eq t -> [RuleResult eq t]) -- Funktion, die auf zwei Termen operiert und ggf. Lsung liefert
    -> Rule es eq t tp                 -- erstellte Regel
cSimpleRuleDet n f = let f' eset eq = let fresult = f eset eq
                                     in  case (fresult) of
						[]        -> [cBacktrackingNo]
						rrs       -> map cBacktrackingDet fresult
                   in  cSimpleRule n f'

