module Gersdorf.ContextMatching.General.Rules.BottomDecompose (
    bottomDecomposeRule 
) where

import Debug.Trace
import qualified Gersdorf.Data.Set as Set

import Gersdorf.ContextMatching.General.Rule
import Gersdorf.ContextMatching.Base.PositionTerm
import Gersdorf.Data.BacktrackingChoice

import Data.Maybe

-- | Regel wie folgt:
--
-- bottomDecompose: 
-- { C[_^k, f(s_1, .., s_n), _^l] =? D[_^k, f(t_1, .., t_n), _^l] } 
--                  -> { s_1 =? t_1, .., s_n =? t_n, C[_^(k+l+1)] =? D[_^(k+l+1)] }
--   (falls s_j=_i <=> t_j=_i, 1 <= j <= n)
--
-- bottomCollision:
-- { C[_^k1, f(s_1, .., s_m), _^l1] =? D[_^k2, g(t_1, .., t_n), _^l2] } 
--                               -> Bot
--   (falls s_j1=_i und t_j2=_i (1 <= j1 <= m, 1 <= j2 <= n) (d. h. mind. ein Loch in Argumenten ==> fixiert Funktionssymbol)
--      und (f != g oder j1 != j2)  (d.h. ungleiches Funktionssymbol oder Loch in versch. Argumenten))
bottomDecomposeRule :: (EnvPosEqSet es eq t tp) => Rule es eq t tp
bottomDecomposeRule = cSimpleRuleDet "Bottom Decompose/Collision" bottomDecompose

bottomDecompose :: (EnvPosEqSet es eq t tp) => 
        es eq t 
        -> eq t 
        -> [RuleResult eq t]
bottomDecompose eset  eq
        | isApplicable
          
          = let -- Paare aller Lcher. Anzahl ist gleich w/ isApplicable
                holePairs = zip (getHolesDeep l) (getHolesDeep r)
                -- Filtert aus allen Lochpaaren die heraus, bei die Vter beider
                -- Lcher Funktionen sind.
                onlyFunctionHolePairs = 
                    filter (\(lh, rh) -> (isFatherFunction lh) && (isFatherFunction rh)) 
                           holePairs

                -- Prft, ob der Vater des bergebenen Terms h (ein Loch)
                -- eine Funktion ist oder nicht.
                isFatherFunction h = checkFather isFunction h
                
            in  calcRuleResult onlyFunctionHolePairs
         
        | otherwise = [RRNothing]
    
        where
        (l, r) = getPair eq

        isApplicable :: Bool
        isApplicable = let leftHoles = countHoles l
                       in  (leftHoles > 0) && (leftHoles == countHoles r)
                           && (not (isHole l) && not (isHole r))

holeNumbersFlat t = map ((+) 1 . getLeftHoles) (getHolesFlat t)

calcRuleResult holePairs = calcRuleResult' 1 Set.emptySet [] holePairs 
    where
    calcRuleResult' holeNum procSet res [] = res
    calcRuleResult' holeNum procSet res (p@(lh, rh):hss) =
        let (processedHN, result) = calcSingleResult p
            newprocSet = Set.listToSet processedHN

            unionedSet = Set.union procSet newprocSet

        in  if (Set.elementOf holeNum procSet)
                then calcRuleResult' (holeNum+1) procSet res hss
                else calcRuleResult' (holeNum+1) unionedSet (result:res) hss
                       
calcSingleResult :: (AnalysisTerm t, PositionTerm t tp, 
                     TreeExtractor t, TermEquation eq t) => 
    (t, t)              -- ^ Paar von parallelen Lochtermen (korresp. 
                        --   Pos.), die direkt unterhalb eines 
                        --   Funktionssymbols sind
    -> ([Int], RuleResult eq t)
calcSingleResult p@(lh, rh) = 
    let leftF  = fromJust (getFather lh) -- muss vorhanden sein w/ onlyFunctionHolePairs!
        rightF = fromJust (getFather rh) -- muss vorhanden sein w/ onlyFunctionHolePairs!

        leftHN = holeNumbersFlat leftF
        rightHN = holeNumbersFlat rightF

        childEqs = childrenExtract leftF rightF
        eq_C'    = makeContext leftF
        eq_D'    = makeContext rightF
        eq_C'_D' = cEquation eq_C' eq_D'
                
    in  if ((getChildIndex lh == getChildIndex rh)
            && (getType leftF == getType rightF)
            && (holeSignatureFlat leftF == holeSignatureFlat rightF)
            && (leftHN == rightHN))
                then (leftHN, cRuleResult (eq_C'_D':childEqs) [])
                else (leftHN, RRBot)
 
