module Gersdorf.ContextMatching.General.Rules.MultiContextDecompose (
    multiContextDecomposeRule 
) where

import Debug.Trace
import qualified Gersdorf.Data.Set as Set

import Gersdorf.ContextMatching.General.Rule
import Gersdorf.ContextMatching.Base.PositionTerm
import Gersdorf.Data.BacktrackingChoice

import Gersdorf.ContextMatching.Base.TermParser
import Gersdorf.ContextMatching.Base.Terms
import Data.Maybe
import Data.List

-- | Regel wie folgt:
--
-- multiContextDecompose:
-- { C[f( C1[_^k1], .., Cn[_^kn] )] =? D[f( D1[_^k1], .., Dn[_^kn] )] } 
--           -> { C[_] =? D[_], C1[_^k1] =? D1[_^k1], .., Cn[_^k1] =? Dn[_^kn] }
--   (falls \exist i,j: 1<=i,j<=n: i /= j, ki >=1 und kj >= 1 
--     (d.h. keine Lcher links und rechts von f, also alle Lcher unterhalb von f.
--      in mindestens zwei versch. Argumenten von f Lcher.)
--
-- multiContextCollision:
-- { C[f( C1[_^k1], .., Cn[_^km] )] =? D[g( D1[_^l1], .., Dn[_^ln] )] } 
--                               -> Bot
--   (falls 1) \exist i,j: 1<=i,j<=m: i /= j, ki >= 1 und kj >= 1
--          2) \exist i,j: 1<=i,j<=n: i /= j, li >= 1 und lj >= 1
--          3) f /= g oder \exist i: 1<=i<=m: ki /= li)
--
--    Argumente 1, 2, und 3 mssen gleichzeitig gelten. 1+2 sorgen dafr,
--    dass unterhalb von f und g jeweils in mind. zwei verschiedenen
--    Argumenten Lcher sitzen (und damit f und g korrespondieren).
--    Argument 3 sorgt fr den Clash, entweder ber das Symbol oder
--    ber unterschiedliche Lochanzahl in den Argumenten.
--
-- Algorithmus:
--
-- Vorbedingung: gleiche Anzahl Lcher links und rechts.
--
-- 1) Sammle fr beide Seiten alle Funktionsterme, die links
--    und rechts von sich keine Lcher haben und in ihrer
--    Lochsignatur (Anzahl der Lcher in ihren Argumenten) in
--    mindestens zwei Argumenten >= 1 Loch besitzen.
-- 2) Bilde alle Kombinationen (t1, t2) der gesammelten Funktionsterme.
--    ACHTUNG: es kann nur eine einzige solche Kombination geben, da
--             es fr jeden der beiden Terme (l, r) jeweils nur eine
--             Funktion mit den geforderten Eigenschaften geben kann.
--             Diese Funktion muss alle Lcher umfassen und gleichzeitig
--             keine Lcher links und rechts haben. Daher gibt es z. B.
--             bei f(a(_, _), b(_, _)) nur den root term als Term, der 1)
--             erfllt! a(_, _) besitzt zwar zwei Lcher, aber die Lcher
--             von b(_, _) sind rechts davon. Hat man also wie f(..) einen
--             Term, der 1) erfllt, so kann man in keines der Argumente
--             mehr abtauchen, da man in dem Fall mindestens ein Loch links
--             oder rechts von sich hat!
--             ABER: a(_, _) und b(_, _) werden in weiteren Schritten erfat!
-- 3) Prfe jedes (t1, t2) ab:
--       a) falls Funktionsname(t1) /= Funktionsname(t2) => Bot
--       b) falls Lochsignatur(t1)  /= Lochsignatur(t2)  => Bot
--       c) sonst: Ergebnis bilden wie oben beschrieben.
--
-- Die Berechnung der Lochsignaturen wird durch das Caching in den
-- ExtendedTerms beschleunigt.


multiContextDecomposeRule :: (EnvPosEqSet es eq t tp) => Rule es eq t tp
multiContextDecomposeRule = cSimpleRule1 "Multicontext Decompose/Collision" multiContextDecompose

multiContextDecompose :: (EnvPosEqSet es eq t tp) => 
        es eq t 
        -> eq t 
        -> RuleResult eq t
multiContextDecompose _ eq
        | isApplicable
          
          = let scoreCandidate :: (TermEquation eq t, PositionTerm t tp, TreeExtractor t) => 
                    (t, t) -> RuleResult eq t
                scoreCandidate (l, r) =
                    if ((getType l == getType r)
                        && (holeSignatureDeep l == holeSignatureDeep r))
                            then let eq_C_D = cEquation (makeContext l) (makeContext r)
                                     eq_cs  = childrenExtract l r
                                 in  cRuleResult (eq_C_D:eq_cs) []
                            else RRBot

                result = scoreCandidate (fromJust cand1, fromJust cand2)
            in  result
         
        | otherwise                    = RRNothing
    
        where
        (l, r) = getPair eq

        cand1 = doubleHoleFunction l
        cand2 = doubleHoleFunction r
       
        leftH = getHoleCount l
       
        isApplicable :: Bool
        isApplicable = (leftH == getHoleCount r)
                       && (leftH >= 2)
                       && (isJust cand1 && isJust cand2)

                                    
-- |Liefert diejenigen Subterme von t zurck, die
--  1) alle Lcher des Terms t umfassen
--  2) Funktionen sind
--  3) in mindestens zwei Argumenten Lcher besitzen (in beliebiger Tiefe)
--    ACHTUNG: es kann nur einen einzige solchen Term geben, da
--             es jeweils nur eine
--             Funktion mit den geforderten Eigenschaften geben kann.
--             Diese Funktion muss alle Lcher umfassen und gleichzeitig
--             keine Lcher links und rechts haben. Daher gibt es z. B.
--             bei f(a(_, _), b(_, _)) nur den root term als Term, der 1)
--             erfllt! a(_, _) besitzt zwar zwei Lcher, aber die Lcher
--             von b(_, _) sind rechts davon. Hat man also wie f(..) einen
--             Term, der 1) erfllt, so kann man in keines der Argumente
--             mehr abtauchen, da man in dem Fall mindestens ein Loch links
--             oder rechts von sich hat!
doubleHoleFunction :: (Term t) => t -> Maybe t
doubleHoleFunction rt = 
    let ts = filter (\t -> isFunction t && (fst $ numberHolePaths t) >= 2) $ allHolesTerms rt
    in  if (null ts) then Nothing
                     else Just $ head ts

-- |Liefert die Anzahl der direkten Kinder zurck, die Lcher enthalten
--  Bsp: numberHolePaths "f(x)" => 0, "f(_)" => 1, "f(a(_), _)" => 2
--       g(f(a(_), _)) => 0
numberHolePaths :: (Term t) => t -> (Int, [Int])
numberHolePaths t = let signature10 = map signum (holeSignatureDeep t)
                    in  (foldl (+) 0 signature10, signature10)

-- |Liefert alle Subterme eines Terms t, die alle in t enthaltenen Lcher
--  umfassen. D. h. bei f(a(_, g())) werden f(..), a(..) und _ geliefert,
--  nicht aber g(), da links von g() ein Loch steht.
allHolesTerms :: (Term t) => t -> [t]
allHolesTerms t = allHolesTerms' [] t
    where
    allHolesTerms' accum t 
        | isContextVariable t = allHolesTerms' (t:accum) (head $ getChildren t)
        | isFunction t        = 
            let cs                = getChildren t
                (nh, signature10) = numberHolePaths t

                processTermList accum [] = accum
                processTermList accum (c:cs) = 
                    processTermList (allHolesTerms' accum c) cs

                restTerms = 
                    if (nh == 0)
                        then processTermList (t:accum) cs
                        else if (nh == 1)
                                then case (findIndex ((/=) 0) signature10) of
                                        Nothing  -> error "Did not find index with /= 0."
                                        (Just i) -> case (getChild t (i+1)) of
                                                        Nothing -> error "Did not find child."
                                                        (Just c) -> allHolesTerms' (t:accum) c
                                else t:accum
            in  restTerms 
        | otherwise           = t:accum
