module Gersdorf.ContextMatching.General.Rules.Split (
   splitRule,
   cCorrPosProvider,
   CorrPosProvider
) where

import Debug.Trace
import Data.Maybe

import Gersdorf.ContextMatching.General.Rule
import Gersdorf.ContextMatching.Base.PositionTerm
import Gersdorf.Data.BacktrackingChoice

data (Term t, TreePath tp2) => CorrPosProvider t tp2 
    = CorrPosProvider {
          cppName :: String,
          cppBacktracking :: Bool,
          cppFunction :: CorrPosFunction t tp2
      }

newtype (Term t, TreePath tp2) => CorrPosFunction t tp2 = 
    CorrPosFunction ((t,t) -> [(tp2, tp2)])

cCorrPosProvider :: (Term t, TreePath tp2) =>
    String
    -> ((t,t) -> [(tp2, tp2)])
    -> Bool
    -> CorrPosProvider t tp2
cCorrPosProvider name f backtrack = CorrPosProvider { cppName = name, 
                                            cppFunction = CorrPosFunction f,
                                            cppBacktracking = backtrack }

applyCorrPosProvider :: (TermEquation eq t, TreePath tp2) =>
    CorrPosProvider t tp2
    -> eq t
    -> [(tp2, tp2)]
applyCorrPosProvider cpp eq = 
    case (cppFunction cpp) of 
         (CorrPosFunction f) -> f (getPair eq)

-- | Regel wie folgt:
--
-- { s[s'|p_s] =? t[t'|p_t] }  ->  { s[_|p_s] =? t[_|p_t], s' =? t' }
--   ( falls p_s und p_t nichttrivial korrespondierende Positionen sind,
--     d. h. weder die root-Position noch Lochpositionen )
--
--
splitRule :: (EnvPosEqSet es eq t tp, TreePath tp2) => 
    CorrPosProvider t tp2
    -> Rule es eq t tp
splitRule cpp = cSimpleRule ("Split: " ++ (cppName cpp)) (split cpp)

split :: (EnvPosEqSet es eq t tp, TreePath tp2) => 
        CorrPosProvider t tp2
        -> es eq t 
        -> eq t 
        -> [BacktrackingChoice (RuleResult eq t)]
split cpp eset eq
        | isApplicable =
              let cResult :: (TermEquation eq t, PositionTerm t tp, TermBuilder t) => 
                      (t, t) -> RuleResult eq t
                  cResult (eq_s', eq_t') = 
                      let eq_s_Ctx = makeContext eq_s'
                          eq_t_Ctx = makeContext eq_t'
                          eq_t'_extr = extractSubTerm eq_t'
                          eq_s'_extr = extractSubTerm eq_s'
                      in  cRuleResult [cEquation eq_s_Ctx eq_t_Ctx, cEquation eq_s'_extr eq_t'_extr] []

              in  if (cppBacktracking cpp) 
                    then [cBacktrackingOne (map cResult corrTerms)]
                    else map (cBacktrackingDet . cResult) corrTerms
          
        | otherwise    = [cBacktrackingNo]
    
        where
        -- initialization of rule variables
        (l, r) = getPair eq
        corrTerms = getCorrTerms l r (applyCorrPosProvider cpp eq)

        isApplicable :: Bool
        isApplicable = not $ null corrTerms


getCorrTerms :: (PositionTerm t tp, TreePath tp2) =>
    t
    -> t
    -> [(tp2, tp2)]
    -> [(t, t)]
getCorrTerms l r =
    filter (not . isTriviallyCorresponding)
    . map (getCorrTermPair (l, r))
    . filter (\(p1, p2) -> not (isEmptyPath p1 || isEmptyPath p2))
    

-- | Nimmt zwei Terme und Pfade zu korrespondierenden Positionen in diesen Termen.
--   Gibt die Subterme zurck, die an den korrespondierenden Positionen sind.
--   Die Subterme sind noch nicht extracted.
getCorrTermPair :: (PositionTerm t tp, TreePath tp2) =>
     (t, t)        -- ^ linker und rechter Term
     -> (tp2, tp2) -- ^ Pfade korrespondierender Positionen
     -> (t, t)     -- ^ korrespondierende Subterme gem. korrespondierender Positionen
getCorrTermPair (l, r) (lp, rp) =
    let nlp = cNewPath l lp
        nrp = cNewPath r rp
        cNewPath t tp = case (cPath t (toPathIndexList tp)) of
                           Nothing -> error ("Did not find Path " ++ show tp ++ " in Term " ++ show t)
                           (Just ntp) -> ntp
        nlt = getNewTerm l nlp
        nrt = getNewTerm r nrp
        getNewTerm t tp = case (gotoPath t tp) of
                             Nothing -> error ("Did not find subterm with path " ++ show tp ++ " in Term " ++ show t)
                             (Just nt) -> nt

    in  (nlt, nrt)

-- | Prft ein Paar korrespondierender Subterme. Sind diese trivial korrespondierend,
--   so wird True zurckgegeben.
--   Subterme sind trivial korrespondierend, wenn sie Lcher sind oder aber
--   beide die Rootposition des Terms sind.
isTriviallyCorresponding :: (Term t) =>
    (t, t)   -- ^ Paar korrespondierender Subterme
    -> Bool  -- ^ True gdw. das Paar trivial korrespondierend ist.
isTriviallyCorresponding (l, r) =
    (isRoot l && isRoot r)
    || isHole l   -- ein Loch muss mit einem Loch korrespondieren!

