module Gersdorf.ContextMatching.Linear.LinearCM (
    solveLinearCM,
    dynamicTableLinearCM,
    solveLinearCMDecision,
    dynamicTableLinearCMDecision
) where

import Data.Array
import Debug.Trace

import Gersdorf.Algorithms.Dynamic

import Gersdorf.ContextMatching.Base.Terms
import Gersdorf.ContextMatching.Base.HeapTerm
import Gersdorf.ContextMatching.Base.SimpleSubstitution
import Gersdorf.ContextMatching.Base.SimpleTermEquation
import Gersdorf.ContextMatching.Base.SemiSolvedSubstitution

type SubstitutionType = SemiSolvedSubstitution HeapTerm HeapPath

solveLinearCMDecision :: (UnificationProblem up e t, TermEquation e t, Term t, TermBuilder t) =>
    up e t 
    -> Bool
solveLinearCMDecision = generateSolution dynamicTableLinearCMDecision 

solveLinearCM :: (UnificationProblem up e t, TermEquation e t, Term t, TermBuilder t) => 
    up e t 
    -> [Substitution HeapTerm]
solveLinearCM = generate . generateSolution dynamicTableLinearCM

generateSolution :: (UnificationProblem up e t, TermEquation e t, Term t, TermBuilder t) =>
     ((HeapTerm, HeapTerm) -> DynTable (Int, Int) a) 
     -> up e t
     -> a
generateSolution tableF up = dynamicSolve dynTable (1,1)
    where
    dynTable = tableF (toHeapTerm left, toHeapTerm right)
    singleTermEq = toSingleTermEquation up
    left = getLeft singleTermEq 
    right = getRight singleTermEq

------------------------------------
-- Decision Problem
------------------------------------

dynamicTableLinearCMDecision :: 
     (HeapTerm, HeapTerm)
     -> DynTable (Int, Int) Bool
dynamicTableLinearCMDecision (l, r) =
    let n = size l
        m = size r
    in  dynamic (computeLCMDecision l r) ((1, 1), (n, m))

computeLCMDecision :: HeapTerm -> HeapTerm -> DynTable (Int, Int) Bool -> (Int, Int) -> Bool
computeLCMDecision left right dt (i, j) = process leftSub rightSub
    where leftSub = getAt left i
          rightSub = getAt right j

          process :: (AdressableTerm t tp) => t -> t -> Bool
          process t1 t2 | isHole t1 && isHole t2              = True
                        | isVariable t1 {- && not (isContext t2) -} = True
                        | isFunction t1 && isFunction t2 && (getName t1 == getName t2) &&
                          (arity t1 == arity t2)              = subSolution leftSub rightSub
                        | isContextVariable t1                = contextSolution leftSub rightSub
                        | otherwise                           = False
          
          contextSolution :: (AdressableTerm t tp) => t -> t -> Bool
          contextSolution s t = or $ map (\subt -> (sol subt))
                                         (toList t)
            where sol subt = dynamicSolve dt (getAdress contextChild, getAdress subt)
                  contextChild = head $ getChildren s
          
          subSolution :: (AdressableTerm t tp) => t -> t -> Bool
          subSolution father1 father2 = and $ 
                (map ( \(ht1,ht2) -> dynamicSolve dt (getAdress ht1, getAdress ht2))
                     (zip (getChildren father1) (getChildren father2)))

-----------------------------------          
-- Solution Problem
-----------------------------------

dynamicTableLinearCM :: 
     (HeapTerm, HeapTerm)
     -> DynTable (Int, Int) SubstitutionType
dynamicTableLinearCM (l, r) =
    let n = size l
        m = size r
    in  dynamic (computeLCM l r) ((1, 1), (n, m))

computeLCM :: HeapTerm -> HeapTerm -> DynTable (Int, Int) SubstitutionType -> (Int, Int) -> SubstitutionType
computeLCM left right dt (i, j) = 
    {-# SCC "computeLCM" #-}
--    trace ("(" ++ show i ++ ", " ++ show j ++ "): " ++ show leftSub ++ " =? " ++ show rightSub ++ " => " ++ show result) result
    result
    where 
          result = simplify $ process leftSub rightSub 
          leftSub = getAt left i
          rightSub = getAt right j

          process :: HeapTerm -> HeapTerm -> SubstitutionType
          process t1 t2 | isHole t1 && isHole t2              = substTop
                        | isVariable t1 {-&& not (isContext t2)-} = substVar leftSub rightSub
                        | isFunction t1 && isFunction t2 && (getName t1 == getName t2) &&
                          (arity t1 == arity t2)              = subSolution leftSub rightSub
                        | isContextVariable t1                = contextSolution leftSub rightSub
                        | otherwise                           = substBot

          contextSolution :: HeapTerm -> HeapTerm -> SubstitutionType
          contextSolution s t = substOr $ map (\subt -> (sol subt)) (toList t)
            where sol subt = substAnd ((cVar subt):[dynamicSolve dt (getAdress contextChild, getAdress subt)])
                  contextChild = head $ getChildren s
                  cVar subt = substContextVar s t (getAdress subt)
          
          subSolution :: HeapTerm -> HeapTerm -> SubstitutionType
          subSolution father1 father2 = substAnd
                (map ( \(ht1,ht2) -> dynamicSolve dt (getAdress ht1, getAdress ht2))
                     (zip (getChildren father1) (getChildren father2)))



