module Gersdorf.Data.ExtendedTree (
    ExtendedTree,
    module Gersdorf.Data.Tree,
    module Gersdorf.Data.StandardTreePath
) where

import Data.Array

import Gersdorf.Data.Tree
import Gersdorf.ContextMatching.Base.Utilities
import Gersdorf.Data.List
import Gersdorf.Data.StandardTreePath

data ExtendedTree a = ExtendedTree {
    value :: a,
    children :: Array Int (ExtendedTree a),

    -- ab hier: Verwaltungsinformationen!
    father :: Maybe (ExtendedTree a),
    childIndex :: Int,
    path :: StandardTreePath
    } 

-- | Wir mssen die Instanz von Eq selbst definieren, damit es wegen der
-- father-Beziehung zu keiner endlosen Rekursion kommt!
instance (Ord a) => Eq (ExtendedTree a) where
    l == r = (value l == value r)
             && (all (foldPair (==)) (zip (getChildren l) (getChildren r)))

instance (Ord a) => Ord (ExtendedTree a) where
    l < r = lessTree l r
    l <= r = (l < r) || (l == r)

{-
instance (Show a) => Show (ExtendedTree a) where
    showsPrec _ n@(ExtendedTree { value = v, 
                                  father = f, 
                                  children = chs }) r
        = let chList = getChildren n
        
              fatherShow :: String
              fatherShow = case (f) of
                                Nothing -> ""
                                (Just father) -> " father=\"" ++ (show (value father)) ++ "\""
          in case (chList) of
                [] -> showString ("<node value=\"" ++ (show v) ++ "\"" ++ fatherShow ++ "/>") r
                (x:xs) -> showPredList ("<node value=\"" ++ (show v) ++ "\"" ++ fatherShow ++ ">\n") chList "\n" "\n</node>\n" r
-}

instance (Ord a) => Listable (ExtendedTree a) where
    toList = toListTree

instance (Ord a) => Flatable ExtendedTree a where
    flatten = map getContent . toList

instance (Ord a) => Tree (ExtendedTree a) where
    getChildren = elems . children
    arity = snd . bounds . children
    getChild t i  | (i < 1 || i > arity t) = Nothing
                  | otherwise              = Just $ children t ! i

instance Container ExtendedTree a where
    getContent t = value t

instance (Ord a) => ContainerTree ExtendedTree a 

instance (Ord a) => TreeExtractor (ExtendedTree a) where
    extractNode t = cNode (getContent t) (map extractNode (getChildren t))

    replaceChildren t css = cNode (getContent t) css

instance (Ord a) => TreeBuilder ExtendedTree a where
    cNode v cs f  = let node = ExtendedTree {
                                 value = v,
                                 father = case (f) of
                                             Nothing -> Nothing
                                             (Just (father, _)) -> Just father,
                                 children = listArray (1, length cs) 
                                                      (map (\ (etcc, n) -> etcc (Just (node, n))) (zip cs [1..])),
                                 childIndex = case (f) of
                                                Nothing         -> 0
                                                (Just (_, nch)) -> nch,
                                 path = case (f) of
                                           Nothing              -> mkPath []
                                           (Just (father, nch)) -> mkPath ((toPathIndexList $ getPath father) ++ [nch])
                                 }
                   in node

instance (Ord a) => UpdatableTree (ExtendedTree a) StandardTreePath where
    transformTreePath f r tp = cTree $ trfTreePath' r tp 0
        where
        trfTreePath' cur tp dp= let mbNewNode = f cur dp
                                in  case (mbNewNode) of
                                      Nothing ->
                                        if (isEmptyPath tp) then
                                            extractNode cur 
                                        else
                                            let nextChild = children cur ! nextChildIdx
                                                nextChildIdx = topIndex tp
                                                newChildrenL = mapNth (\_ -> trfTreePath' nextChild (popPath tp) (dp + 1))
                                                                      (\c -> (extractNode c))
                                                                      nextChildIdx
                                                                      (getChildren cur)
                                            in cNode (getContent cur) newChildrenL
                                                             
                                      (Just newNode) -> newNode
                                        
                                        
                      

    replaceSubtree ct nnode 
        | isRoot ct = cTree nnode -- falls der ChildTree bereits root war -> nct wird auch root
        | otherwise = transformTreePath (replaceLeaf nnode) (getRoot ct) ctPath

        where
        ctPath = getPath ct
        replaceLeaf nnode curnode depth = if (depth == pathDepth ctPath) then Just nnode
                                                                         else Nothing

instance (Ord a) => NavigatableTree (ExtendedTree a) where
    getFather t = father t


instance (Ord a) => AdressableTree (ExtendedTree a) StandardTreePath where
    cPath t itp = let tp = mkPath itp
                  in  case (gotoPath t tp) of
                        Nothing  -> Nothing
                        (Just _) -> Just tp

    getPath t = path t

    getChildIndex t = childIndex t

instance (Ord a) => ContainerNavigatableTree ExtendedTree a 

instance (Ord a) => ContainerAdressableTree ExtendedTree a StandardTreePath

instance (Ord e, ContainerTree t1 e) => ConvertableContainerTree t1 ExtendedTree e where
    convertTree = convertTree'
        where 
        convertTree' t1 = let v  = getContent t1
                              cs = getChildren t1
                          in  cNode v (map convertTree' cs)

