-- | Das Modul beinhaltet die Implementierung fr HeapTrees. Ein HeapTree ist ein
-- Baum, dessen Implementierung in einem Heap liegt.
module Gersdorf.Data.HeapTree (
    HeapAddr,
    HeapTree,
    HeapCell,
    HeapPath,
    toHeapTree,
    module Gersdorf.Data.Tree
) where

import Gersdorf.Data.Tree
import Gersdorf.Data.SimpleTree
import Data.Array
import Gersdorf.ContextMatching.Base.Utilities

-- | Die Adressen des Heaps sind Integers.
type HeapAddr = Int

-- | Eine Heapzelle beinhaltet ein Element des Heaps. Der Datentyp ist polymorph ber
-- dem Elementtyp a.
data HeapCell a 
    = HeapCell (Maybe HeapAddr) -- ^ evtl. Zeiger auf Vater
                                --   (Die Wurzel besitzt keinen Vater -> Nothing)
               a                -- ^ Knoteninhalt
               [HeapAddr]       -- ^ Zeiger auf Kinder
                                                

-- | Ein HeapTree ist ein Array von Heapzellen gemeinsam mit einem Zeiger auf die Zelle des
-- aktuellen Knotens.
data HeapTree a 
    = HeapTree HeapAddr         -- ^ Zeiger auf Baumwurzelknoten
               (Array HeapAddr (HeapCell a)) -- ^ Array mit Baumknoten

-- | Ein HeapPath beinhaltet lediglich die Adresse des Heaps. Sie wird zu einem echten Path
-- erst durch Auswertung
data HeapPath = HeapPath HeapAddr [Int] -- zweites Argument: der Path, lazily ausgewertet!
    deriving (Eq, Show)

{-
instance (Show a) => Show (HeapTree a) where
    showsPrec _ n r
        = let chList = getChildren n
              fatherM = getFather n
              value = getContent n
        
              fatherShow :: String
              fatherShow = case (fatherM) of
                                Nothing -> ""
                                (Just father) -> " father=\"" ++ (show (getContent father)) ++ "\""
          in case (chList) of
                [] -> showString ("<node value=\"" ++ (show value) ++ "\"" ++ fatherShow ++ "/>") r
                (x:xs) -> showPredList ("<node value=\"" ++ (show value) ++ "\"" ++ fatherShow ++ ">\n") chList "\n" "\n</node>\n" r
-}

-- | Zwei HeapTrees knnen auf Gleichheit verglichen werden, falls der Elementtyp in Eq ist.
instance (Eq a) => Eq (HeapTree a) where
    -- Der Vergleich erfolgt ber die Struktur des Terms.
    (HeapTree lr la) == (HeapTree rr ra) = equalCell (la ! lr) (ra ! rr)
        where
        equalCell :: HeapCell a -> HeapCell a -> Bool
        equalCell (HeapCell _ lval lcs) (HeapCell _ rval rcs) =
            (length lcs == length rcs) && (lval == rval) &&
            (all (\(laddr, raddr) -> equalCell (la ! laddr) (ra ! raddr)) $ zip lcs rcs)

instance (Ord a) => Ord (HeapTree a) where
    l < r = lessTree l r
    l <= r = (l < r) || (l == r)

-- | Ein HeapTree ist Listable. Dabei wird werden die Subtrees aufgelistet.
instance (Ord a) => Listable (HeapTree a) where
    toList = toListTree 

    size (HeapTree _ arr) = let (_, n) = bounds arr
                            in  n

-- | Ein HeapTree ist Flatable. Dabei werden die Elemente aufgelistet.
instance (Ord a) => Flatable HeapTree a where
    -- Benutze die Eigenschaft als Flatable und hole die Elemente.
    flatten = map getContent . toList

-- | Ein HeapTree ist ein Container. 
instance Container HeapTree a where
    -- liefer den Inhalt des Wurzelknotens.
    getContent (HeapTree root arr) = case (arr ! root) of
                                        (HeapCell _ val _) -> val 

-- | Ein HeapTree ist ein Tree.
instance (Ord a) => Tree (HeapTree a) where
    -- liefere die Kinder der Wurzel als HeapTrees.
    getChildren (HeapTree root arr) = case (arr ! root) of
                                        (HeapCell _ _ chs) -> map (\a -> HeapTree a arr) chs

instance (Ord a) => NavigatableTree (HeapTree a) where
    -- Liefer den Vater des aktuellen Knotens, es sei denn, dieser existiert nicht. In diesem
    -- Falle handelt es sich um die Wurzel des Gesamtbaums.
    getFather (HeapTree root arr) = case (arr ! root) of
                                        (HeapCell Nothing _ _)      -> Nothing
                                        (HeapCell (Just faddr) _ _) -> Just $ HeapTree faddr arr

    -- Liefert die Wurzel des Gesamtbaums, ausgehend vom aktuellen Knoten. Sind wir
    -- bereits an der Wurzel, so wird diese trotzdem geliefert.
    getRoot t@(HeapTree root arr) = case (arr ! root) of
                                        (HeapCell Nothing _ _)  -> t
                                        (HeapCell (Just _) _ _) -> case (getFather t) of
                                                                      Nothing -> t
                                                                      (Just f) -> getRoot f

    -- Wir prfen, ob der aktuelle Knoten die Wurzel des Baums ist.
    isRoot (HeapTree root arr) = case (arr ! root) of
                                    (HeapCell Nothing _ _) -> True
                                    _                      -> False

instance Adressable (HeapTree a) where
   -- Liefert die Adresse des aktuellen Knotens.
   getAdress (HeapTree root _) = root

instance AdressTarget (HeapTree a) where
   -- Wir liefern einen Baum, dessen aktueller Knoten die gewnschte Adresse ist.
   -- ACHTUNG: Eine berprfung auf Zulssigkeit der Adresse findet nicht statt!
   getAt (HeapTree _ arr) addr = HeapTree addr arr

instance AdressableContainer HeapTree a 

instance (Ord a) => ContainerTree HeapTree a 

instance (Ord a) => ContainerNavigatableTree HeapTree a 

instance (Ord a) => AdressableTree (HeapTree a) HeapPath where
    cPath t itp = case (cPath' (getRoot t) itp) of
                    Nothing  -> Nothing
                    (Just t) -> Just (HeapPath (getAdress t) itp)
        where cPath' t []     = Just t
              cPath' t (i:is) = case (getChild t i) of
                                    Nothing  -> Nothing
                                    (Just c) -> cPath' c is
    
    -- schnellere O(1)-Implementierung als die Standard-Implementierung in der Typklasse!
    gotoPath t (HeapPath addr _)
        | addr < 1 || addr > size t = Nothing
        | otherwise                 = Just $ getAt t addr
                                    
    getPath t@(HeapTree targetAddress _) = HeapPath targetAddress (getPath' t [])
        where
        getPath' :: HeapTree a -> [Int] -> [Int]
        getPath' t@(HeapTree cellAddress _) is 
            | isRoot t  = is
            | otherwise = case (getFather t) of
                            Nothing -> error "No father found. may not happen!"
                            Just f  -> getPath' f ((getChildIndex f cellAddress):is)
        
        getChildIndex :: HeapTree a -> HeapAddr -> Int
        getChildIndex f a =  getChildIndex' (getChildren f) a 1
            where
            getChildIndex' [] targetAddr _                     = 0
            getChildIndex' ((HeapTree r _):cs) targetAddr i 
                | r == targetAddr = i
                | otherwise       = getChildIndex' cs targetAddr (i+1)

    -- |Nimmt den ersten HeapTerm und setzt an HeapPosition ein Loch ein. 
    -- Achtung: hier primitive Lsung. Kein Inplace-Update! Daher O(n)!
{-    replace (HeapTree root arr) pos val = HeapTree root (arr // [(pos, prepareCell (arr ! pos))])
        where
        prepareCell (HeapCell f _ _) = HeapCell f val []
-}

instance (Ord a) => ContainerAdressableTree HeapTree a HeapPath

-- |Convert a Tree into a HeapTree. 
-- A HeapTree consists of an Array containing HeapCells as defined above. In addition it
-- contains a root pointer pointing to the rootnode of the Tree.
-- Each HeapCell may point to one or more other HeapCells via HeapAddr. A HeapCell
-- may have a father.
toHeapTree :: (ContainerTree t a) => t a -> HeapTree a
toHeapTree t = HeapTree 1 (array (1,n) (snd (transformTree t 1 Nothing)))
               where 
               n = size t
               -- This helper function transforms the given SimpleTerm to the heap at the
               -- given Starting heap address. All subterms will be processed as well.
               -- It returns a pair of the next free heap address to be used and a list
               -- of associations of HeapAddrs to HeapCells containing the Subterms.
               transformTree :: (ContainerTree t a) => t a -- aktueller Node
                  -> HeapAddr                -- nchster freier Slot
                  -> Maybe HeapAddr          -- Vaternode
                  -> (HeapAddr, [(HeapAddr, HeapCell a)])
               transformTree t cur father = (nextcur, (cur, HeapCell father (getContent t) childrenRoots):transformedChildren)
                    where 
                    --transformedNode :: [(HeapAddr, [(HeapAddr, HeapCell a)])]
                    -- transforms the list of children, i. e. each subtree is put into the list
                    -- The first component of each
                    -- list item represents the heap address of the argument (i. e. the children's
                    -- root). The second component contains all HeapCells (and it's associations)
                    -- of all the transformed subtrees of the child.
                    transformedNode = transformNodeList (getChildren t) (cur+1) (Just cur)
                    -- the root addresses of the transformed function arguments
                    childrenRoots :: [HeapAddr]
                    childrenRoots = fst (unzip transformedNode)
                    -- the associations of HeapAddrs to HeapCells containing the children's transformed
                    -- subterms. This list contains all HeapCells of _all_ children.
                    --transformedChildren :: [(HeapAddr, HeapCell a)]
                    transformedChildren = concat $ snd (unzip transformedNode)
                    -- the next Heap Address to be used. This is the address of the last
                    -- cell of the Children-Cells plus one.
                    nextcur :: HeapAddr
                    nextcur = case transformedChildren of 
                                l:ls -> (fst (last transformedChildren)) + 1
                                [] -> cur+1

                    -- Helper function that transforms a list of SimpleTerms into an association of
                    -- HeapAddrs and HeapCells. Additionally for each given SimpleTerm its root address
                    -- is returned in the first component of each list item.
                    transformNodeList :: (ContainerTree t a) => [t a] -- Liste der Nodes
                        -> HeapAddr -- nchster freier Slot
                        -> Maybe HeapAddr -- Vaternode
                        -> [(HeapAddr, [(HeapAddr, HeapCell a)])]
                    transformNodeList (t:ts) cur father = (cur, transformedTree):(transformNodeList ts nextcur father)
                       where (nextcur, transformedTree) = transformTree t cur father
                    transformNodeList [] cur father = []

instance (Ord e) => TreeExtractor (HeapTree e) where
    extractNode t = (\_ -> extractNode' t)
        where
        extractNode' = toHeapTree  -- toHeapTree beginnt beim root des trees.

    replaceChildren t css = 
        let stChilds = (map (convertTree . cTree) css) :: [TreeNode (SimpleTree e)]
            newTree = extractNode $ toHeapTree $ cTree $ cNode (getContent t) stChilds
        in  extractNode $ replaceSubtree t newTree

instance (Ord e) => UpdatableTree (HeapTree e) HeapPath where
    -- | Implementiert durch Traversierung und Umbau in einen SimpleTree
    -- Resultat: O(n)
    replaceSubtree ct@(HeapTree targetaddr arr) nnode = 
        let ntree = convertTree (cTree nnode) :: TreeNode (SimpleTree e)
            root  = getRoot ct
    
            traverse :: HeapAddr -> TreeNode (SimpleTree e) -> HeapTree e -> TreeNode (SimpleTree e)
            traverse targetaddr ntree curNode
                | (getAdress curNode) == targetaddr = ntree
                | otherwise                         =
                    cNode (getContent curNode)
                          (map (traverse targetaddr ntree) (getChildren curNode))
           
        in  toHeapTree (cTree $ traverse targetaddr ntree root)

    transformTreePath f t tp = transform f 0 (getPathNodes t tp)
        where transform f d []     = getRoot t
              transform f d (t:ts) = case (f t d) of
                                        Nothing   -> transform f (d+1) ts
                                        (Just tn) -> replaceSubtree t tn
    

instance TreePath HeapPath where
    toPathIndexList (HeapPath _ l) = l
   
    topIndex (HeapPath _ []) = error "no top index"
    topIndex (HeapPath _ (x:_)) = x
    
    popPath (HeapPath _ []) = error "no top index"
    popPath (HeapPath a (_:xs)) = HeapPath a xs

    isEmptyPath (HeapPath _ []) = True
    isEmptyPath _ = False

instance Adressable HeapPath where
    getAdress (HeapPath a _) = a

instance (Ord e, ContainerTree t1 e) => ConvertableContainerTree t1 HeapTree e where
    convertTree t1 = (\_ -> toHeapTree t1)
