module Gersdorf.Data.List (
    predSplit,
    removeEach,
    replaceNth,
    mapNth,
    mapRemove,
    foldPair,
    uniques, uniquesBy,
    zipEquals, zipEqualsBy,
    FunctorRemove(..),
    countOccurences,
    joinEquals,
    calcAverage
) where

import Data.List
import qualified Data.FiniteMap as FM

-- |Nimmt ein Prdikat und eine Liste; liefert zwei Listen
-- zurck: eine, mit den positiven Werten, die andere mit den
-- negativen Werten gem. Prdikat. Durchluft die Ursprungsliste
-- nur einmal --> O(n).
predSplit :: (a -> Bool) -> [a] -> ([a], [a])
predSplit pred l = predSplit' pred l ([], [])
    where
    predSplit' pred [] res = res
    predSplit' pred (l:ls) (t, f) = 
        case (pred l) of
            True -> predSplit' pred ls (l:t, f)
            False -> predSplit' pred ls (t, l:f)

-- |Nimmt eine Liste, lscht jedes Element der Liste jeweils einmal
-- und gibt eine Liste von Paaren zurck, in der das gelschte Element
-- sowie die Restliste aufgefhrt sind. Reihenfolge ist nicht garantiert!
-- Bsp: removeEach [1,2,3] = [([2,3],1), ([1,3],2), ([1,2], 3)]
removeEach :: [a] -> [([a], a)]
removeEach l = removeEach' [] l
    where
    removeEach' pre []     = []
    removeEach' pre (x:xs) = (pre ++ xs, x):(removeEach' (x:pre) xs)

mapRemove :: (a -> Maybe b) -> [a] -> [b]
mapRemove f []     = []
mapRemove f (x:xs) = case (f x) of
                        Nothing -> mapRemove f xs
                        (Just nx) -> nx:(mapRemove f xs)

replaceNth :: (Int, a) -> [a] -> [a]
replaceNth n l = replaceNth' n l 1
    where
    replaceNth' _ [] _ = []
    replaceNth' (index, nv) (x:xs) i | index == i = nv:xs
                                     | otherwise  = x:(replaceNth' n xs (i+1))

mapNumber :: (Int -> a -> b) -> [a] -> [b]
mapNumber f l = mapNumber' f 1 l
    where
    mapNumber' f i [] = []
    mapNumber' f i (x:xs) = (f i x):(mapNumber' f (i+1) xs)

mapNth :: (a -> b) -> (a -> b) -> Int -> [a] -> [b]
mapNth nFunc othFunc n l = 
    let f i x = if (i == n) then nFunc x
                            else othFunc x
    in  mapNumber f l

class (Functor f) => FunctorRemove f where
	fmapRemove :: (a -> Maybe b) -> f a -> f b

instance FunctorRemove [] where
	fmapRemove = mapRemove

foldPair :: (a -> b -> c)
    -> (a, b)
    -> c
foldPair f (a, b) = f a b

-- | gibt diejenigen Elemente der Liste zurck, die 
--   genau einmal vorkommen.
--   Sortiert die Liste und schaut dann nach.
--   Komplexitt: O(n) = O(lg n) + O(n)
uniques :: (Ord a) => [a] -> [a]
uniques as = uniquesBy compare as
    
uniquesBy :: (a -> a -> Ordering) -> [a] -> [a]
uniquesBy cmpF as = 
    concat 
    $ filter justOne 
    $ groupBy (\a b -> cmpF a b == EQ) 
    $ sortBy cmpF as
    where
    -- | gibt zurck, ob die Liste genau ein Element enthlt. Besser als length l == 1,
    --   da hier nicht die komplette Liste ausgewertet werden muss, sondern nur das
    --   erste Listenelement.
    justOne :: [a] -> Bool
    justOne [e] = True
    justOne _   = False


-- | nimmt zwei Listen und zipped gleiche Elemente.
-- Voraussetzungen: 1) in jeder Liste kommt jedes Element hchsten einmal vor
--                     (siehe uniques)
--                  2) beide Listen sind sortiert.
--
-- Bsp: zipEquals [1,2,3] [1,3] -> [(1,1),(3,3)]
--      zipEquals [] [1,3]      -> []
--      zipEquals [1,2] [1,3]   -> [(1,1)]
zipEquals :: (Ord a) => [a] -> [a] -> [(a, a)]
zipEquals as bs = zipEqualsBy compare as bs

zipEqualsBy :: (a -> a -> Ordering) -> [a] -> [a] -> [(a, a)]
zipEqualsBy _ [] _ = []
zipEqualsBy _ _ [] = []
zipEqualsBy cmpF la@(a:as) lb@(b:bs) 
    | (cmpF a b == EQ) = (a,b):(zipEqualsBy cmpF as bs)
    | (cmpF a b == LT) = zipEqualsBy cmpF as lb
    | otherwise = zipEqualsBy cmpF la bs


-- | Zhlt die Anzahl Vorkommen von Elementen in einer Liste und gibt
--   die unterschiedlichen Elemente der Liste zurck gemeinsam mit
--   der Anzahl, in der die in der Ursprungsliste vorkommen.
--   Bsp: countOccurences [1,1,2,3,2,4,6,6]
--          -> [(1,2),(2,2),(3,1),(4,1),(6,2)]
countOccurences :: (Ord a) => 
    [a]              -- ^ Eingabeliste
    -> [(a, Int)]    -- ^ Liste von Paaren (Element, Anzahl Vorkommen in Ursprungsliste)
countOccurences as =
    joinEquals (zip as (repeat 1)) (+)

joinEquals :: (Ord a) =>
    [(a, b)]
    -> (b -> b -> b)
    -> [(a, b)]
joinEquals maps aggregate =
    let countMap = FM.addListToFM_C (\i1 i2 -> aggregate i1 i2) FM.emptyFM maps
    in  FM.fmToList countMap

-- | Berechnet den durchschnittlichen Wert aller Werte der Liste
calcAverage :: (Integral a) => [a] -> Double
calcAverage ls =
    let numEl   = fromInteger $ toInteger $ length ls
        totalEl = fromInteger $ toInteger $ foldl (+) 0 ls
    in  (totalEl / numEl)

    
