module Gersdorf.Data.Tree (
    Tree(..),
    NavigatableTree(..),
    ContentFunction,
    ContainerTree,
    ContainerNavigatableTree,
    ContextableTree(..),
    ContextTreeBuilder(..),
    AdressableTree(..),
    ContainerAdressableTree,
    TreeExtractor(..),
    TreeBuilder(..),
    UpdatableTree(..),
    TreePath(..),
    TreePathBuilder(..),
    ConvertableContainerTree(..),
    TreeNode,
    transformTree,
    transformTreeState,
    toListTree,
    lessTree,
    TransformStateFunction,
    TransformAction(..),
    path2MonadicTree, path2MonadicTreeC, 
    module Gersdorf.Data.Listable,
    module Gersdorf.Data.Container,
    module Gersdorf.Data.Flatable
) where

import Gersdorf.Data.Listable
import Gersdorf.Data.Container
import Gersdorf.Data.Flatable
import Gersdorf.Data.List

class (Show tp, Eq tp) => TreePath tp where
    toPathIndexList :: tp -> [Int]
    topIndex :: tp -> Int
    popPath  :: tp -> tp
    isEmptyPath :: tp -> Bool
    pathDepth :: tp -> Int
    pathDepth = length . toPathIndexList

class (TreePath tp) => TreePathBuilder tp where
    mkPath :: [Int] -> tp

class (Listable t, Ord t) => Tree t where
    getChildren :: t -> [t]

    getChild :: t -> Int -> Maybe t
    getChild t i  | (i < 1 || i > arity t) = Nothing
                  | otherwise              = Just $ getChildren t !! (i - 1)

    -- | Number of children
    arity :: t -> Int
    arity = length . getChildren

    -- | gibt alle Bltter des Baums in einer Liste zurck
    leaves :: t -> [t]
    leaves = filter (\t -> arity t == 0) . toList
    

class (Tree t) => NavigatableTree t where
    getFather :: t -> Maybe t

    getRoot :: t -> t
    getRoot t = case (getFather t) of
                    Nothing -> t
                    (Just f) -> getRoot f
    isRoot :: t -> Bool
    isRoot t = case (getFather t) of
                    Nothing -> True
                    (Just _) -> False

    checkFather :: (t -> Bool) -> t -> Bool
    checkFather p t =
        case (getFather t) of
            Nothing  -> False
            (Just f) -> p f

newtype (ContainerTree t e) =>
  ContentFunction t e = ContentFunction (t e -> e)

class (Tree (t e), Container t e, Flatable t e, Eq e) => ContainerTree t e 
    
class (ContainerTree (t c) e) => ContextableTree t c e  where
    getContext :: t c e  -> c
    getContextFunction :: t c e -> (t c e -> c)

class (NavigatableTree (t e), ContainerTree t e) => ContainerNavigatableTree t e

class (ContainerNavigatableTree t e, AdressableTree (t e) tp) => ContainerAdressableTree t e tp

class (NavigatableTree t, TreePath tp) => AdressableTree t tp | t -> tp where
    cPath :: t -> [Int] -> Maybe tp

    getPath :: t -> tp
    
    gotoPath :: t -> tp -> Maybe t
    gotoPath t tp | isEmptyPath tp = Just t
                  | otherwise      = case (topChild t tp) of
                                        Nothing -> Nothing
                                        (Just c) -> gotoPath c (popPath tp)

    -- | Gibt das erste Kind entsprechend des Paths zurck. Falls der Path leer ist,
    -- wird der Knoten selbst zurckgegeben.
    topChild :: t -> tp -> Maybe t
    topChild t tp | isEmptyPath tp = Just t              
                  | otherwise      = getChild t (topIndex tp)

    -- | Returns the Treenodes along the given tree path
    getPathNodes :: t -> tp -> [t]
    getPathNodes t tp | isEmptyPath tp = [t]
                      | otherwise      = let nextChildM = topChild t tp
                                         in  case (nextChildM) of
                                                Nothing  -> []
                                                (Just c) -> t:(getPathNodes c (popPath tp))


    
    -- Return the index number as a child of this node. if the node is the root node,
    -- return 0.
    getChildIndex :: t -> Int
    getChildIndex t | isEmptyPath tp = 0
                    | otherwise      = last $ toPathIndexList tp
        where tp = getPath t

    getTreeNodeArgs :: t
        -> TreeNodeArgs t
    getTreeNodeArgs t = case (getFather t) of
                            Nothing  -> Nothing
                            (Just f) -> Just (f, getChildIndex t)

type TreeNode tb = TreeNodeArgs tb      
    -> tb                                -- ^ created Tree (Tree t over a)
type TreeNodeArgs tb = Maybe (tb, Int)   -- ^ father and position of this child

class (Tree t) => TreeExtractor t where
    -- | Converts a Node to a tree by using the given node as the root node.
    cTree :: TreeNode t                   -- ^ the root node
        -> t                              -- ^ the resulting tree
    cTree tn = tn Nothing

    -- | may take O(n) because of processing of the tree path
    extractNode :: t                       -- ^ the tree
        -> TreeNode t                      -- ^ the resulting tree node

    replaceChildren :: t
        -> [TreeNode t]
        -> TreeNode t

class (TreeExtractor (t e), ContainerTree t e) => TreeBuilder t e where
    -- | Creates a node of the tree. This won't be a Tree itself, but
    --   must be converted to a tree via the Function @cTree@
    cNode :: e                             -- ^ content
        -> [TreeNode (t e)]                -- ^ children
        -> TreeNode (t e)                  -- ^ new node

class (TreeExtractor (t c e), ContextableTree t c e) => ContextTreeBuilder t c e where
    cNodeC :: e
        -> (t c e -> c)
        -> [TreeNode (t c e)]
        -> TreeNode (t c e)

class (AdressableTree t tp) => UpdatableTree t tp where
    -- | replaces a subtree t1 with a subtree t2 and returns the
    -- modified subtree t2 with correct subtree-relationships
    replaceSubtree :: 
        t                    -- ^ the subtree t1 to delete
        -> TreeNode t        -- ^ the subtree node t2 to put at the place of t1
        -> t                 -- ^  the new root tree (old root of t1, new root of t2)

    replacePath ::             
        t                    -- ^ Baum, in dem Ersetzung vorzunehmen ist
        -> tp                -- ^ Pfad zum Zielknoten
        -> TreeNode t        -- ^ Knoten, durch den der Zielknoten ersetzt werden soll
        -> Maybe t           -- ^ root des Ergebnisbaums.
    replacePath t tp tn = 
        do targetNode <- gotoPath t tp
           return $ replaceSubtree targetNode tn

    transformTreePath ::  
        (t                           -- ^ Aktuelle Baumknoten
            -> Int                   -- ^ Tiefe, die bisher abgelaufen wurde
            -> Maybe (TreeNode t))   -- ^ Nothing, falls keine Aktion, d. h. der Knoten soll erhalten
                                     --   bleiben. Sonst: Node, der den aktuellen Baumknoten ersetzen soll.
        -> t                     -- ^ Wurzelknoten, von dem aus der Pfad geht
        -> tp                    -- ^ Pfad, der abzulaufen ist
        -> t                     -- ^ Ergebnis: Wurzel des neuen Baums

type TransformStateFunction s t = s -> t -> (s, TransformAction t)
data TransformAction t = 
    TANothing                       -- ^ Knoten bleibt unangetastet
    | TACompleteTree (TreeNode t)   -- ^ Knoten wird durch einen neuen Baum ersetzt
    | TAThisNode ([TreeNode t] -> TreeNode t) -- ^ Funktion, die den neuen Knoten baut.

-- |transform the tree. 
-- The transformation function of type @t -> Maybe t@ needs some explanation.
-- If this function returns @Nothing@, then the Tree won't be changed and
-- map descends deeper into the tree. If returning Just t, then the old
-- tree will be replaced by t.

transformTree :: (TreeExtractor t, AdressableTree t tp) =>
    (t -> TransformAction t) -- ^The function Transforming a Tree t into a Tree t'
    -> t                -- ^The source Tree
    -> t                -- ^The resulting Tree 
transformTree f t = snd $ transformTreeState f' () t
    where
    f' _ t = ((), f t)

-- | TODO: evtl. mit StateMonad neu schreiben; knnte wesentlich einfacher sein.
transformTreeState :: (TreeExtractor t, AdressableTree t tp) =>
    TransformStateFunction s t     -- Transformationsfunktion mit Status
    -> s                              -- Startstatus
    -> t                              -- Startknoten
    -> (s, t)                        -- finaler Status und Ergebnis
transformTreeState f startstate t = 
    let descendMap f curstate tn tnargs css = 
            let (newstate, ncss) = processChildren curstate css
            in  (newstate, ncss)

        processChildren state [] = (state, [])
        processChildren state (c:cs) = 
            let (state', nc)      = transformer' f state c
                (laststate, ocss) = processChildren state' cs
            in  (laststate, nc:ocss)
            
        transformer' f curstate t = 
            let (newstate, action_t) = f curstate t
                (newstate', newcss)  = processChildren newstate (getChildren t)
                tnargs               = getTreeNodeArgs t
                
                (finalstate, newnode) = 
                    case (action_t) of
                        (TANothing)          -> (newstate', replaceChildren t newcss)
                        (TACompleteTree tn') -> transformer' f newstate (tn' tnargs)
                        (TAThisNode tnf')    -> (newstate', tnf' newcss) 
                        
            in  (newstate', newnode)

        (finalstate, finalresult) = transformer' f startstate t

    in  (finalstate, cTree finalresult)
                         


class (ContainerTree t1 e, ContainerTree t2 e) => ConvertableContainerTree t1 t2 e where
    convertTree :: t1 e -> TreeNode (t2 e)

-- Wir knnen leider nicht einfach schreiben
--    instance (Tree t) => Listable t where ...
-- da wir damit in undecidable instances laufen und -fallow-undecidable-instances
-- nutzen mssten. Deshalb eine etwas umstndlichere Formulierung.
toListTree :: (Tree t) => t -> [t]
toListTree t = toListAccum [] t
    where
    toListAccum accum t 
        | (arity t == 0) = t:accum
        | otherwise      = toListAccumL (t:accum) (getChildren t)
       
    toListAccumL accum []     = accum
    toListAccumL accum (t:ts) = toListAccumL (toListAccum accum t) ts

    {- ursprngliche und wesentlich ineffizientere Version (w/ concat) 
    toList n = n:(convertList (getChildren n))
        where
        convertList :: (Tree t) => [t] -> [t]
        convertList = concat . map toList
    -}

-- W/ undecidable instances ist dies eine Hilfsmethode, da wir nicht einfach
--    instance (ContainerTree t e, Ord e) => Ord (t e) where ...
-- schreiben knnen, die Funktion jedoch in allen Treetypen gleich aussieht.
--
-- Prft, ob Baum l kleiner als Baum r ist.
lessTree :: (ContainerTree t e, Ord (t e), Ord e) => t e -> t e-> Bool
lessTree l r = (getContent l < getContent r) 
               || (arity l < arity r)
               || checkLexico (zip (getChildren l) (getChildren r))
   where
   checkLexico [] = False
   checkLexico ((l,r):ts) 
        | l == r = checkLexico ts
        | l <  r = True
        | otherwise = False

-- | Nimmt einen Baum und einen TreePath. Gibt einen neuen Baum zurck,
--   der ein als Liste entarteter Baum mit den Knoteninhalten auf dem 
--   Pfadweg ist.
path2MonadicTree :: (TreeBuilder t e, AdressableTree (t e) tp) => 
    t e
    -> tp 
    -> t e
path2MonadicTree t tp = 
    cTree $ path2MonadicTreeHelper nodeCreatorF t tp
    where
    nodeCreatorF :: (TreeBuilder t e) =>
        t e -> [TreeNode (t e)] -> TreeNode (t e)
    nodeCreatorF t cs = cNode (getContent t) cs

-- | Nimmt einen Baum und einen TreePath. Gibt einen neuen Baum zurck,
--   der ein als Liste entarteter Baum mit den Knoteninhalten auf dem 
--   Pfadweg ist.
path2MonadicTreeC :: (ContextTreeBuilder t c e, AdressableTree (t c e) tp) => 
    t c e
    -> tp 
    -> t c e
path2MonadicTreeC t tp =
    cTree $ path2MonadicTreeHelper nodeCreatorF t tp
    where
    nodeCreatorF :: (ContextTreeBuilder t c e) => 
        t c e -> [TreeNode (t c e)] -> TreeNode (t c e)
    nodeCreatorF t cs = cNodeC (getContent t) (getContextFunction t) cs

path2MonadicTreeHelper :: (TreeExtractor t, AdressableTree t tp) => 
    (t -> [TreeNode t] -> TreeNode t)
    -> t
    -> tp 
    -> TreeNode t
path2MonadicTreeHelper nodeCreatorF t tp = 
    nodeCreatorF t (path2MonadicTree' tailNodes)
    where
    pathNodes = getPathNodes t tp
    headNode = head pathNodes
    tailNodes = tail pathNodes
    path2MonadicTree' []     = []
    path2MonadicTree' (t:ts) = [nodeCreatorF t (path2MonadicTree' ts)]
