module Test.Center.Commands (
    CommandDescr(..),
    Commands(..),
    ArgDescr(..),
    executeCommand,
    showHelp,
    stringArgument,
    intArgument,
    liftArgument,
    unknown
) where

import Test.Center.IOHelpers
import GHC.Read (readEither)
import Gersdorf.ContextMatching.Base.Utilities

--
-- Command library
--

data CommandDescr a r = CommandDescr String [ArgDescr a] ([a] -> r)
data Commands a r = Commands [(String, CommandDescr a r)]

data ArgDescr a = Argument String (String -> Either String a)
                  | RestArgument String (String -> a)

executeCommand :: String -> Commands a r -> (String -> String -> r) -> r
executeCommand s (Commands cmds) def = case (lookup getCmd cmds) of
                                          Just (CommandDescr _ argdescr cmd) -> 
                                            case (parseArgs argdescr getArgs) of
                                                Left e -> def s (e ++ " Definition is: " ++ (generateCommandStr getCmd argdescr))
                                                Right as -> cmd as
                                          Nothing -> def s ("Did not understand command.")
    where parseCmds :: [String]
          parseCmds = parseWords s

          getCmd :: String
          getCmd = if (length parseCmds >= 1) then (parseCmds !! 0) else ""

          getArgs :: [String]
          getArgs = case parseCmds of
                       [] -> []
                       [c] -> []
                       (c:aa) -> aa
          
parseWords :: String -> [String]
parseWords s = process (lex s)
    where process :: [(String, String)] -> [String]
          process [] = []
          process ((s, r):rr) | length r > 0 = ((unstring s):process (lex r))
                              | otherwise = [unstring s]
          unstring :: String -> String
          unstring [] = []
          unstring full@(s:ss) = if (s == '\"') then take (length ss - 1) ss
                                                else full
          
parseArgs :: [ArgDescr a] -> [String] -> Either String [a]
parseArgs argDescrs args 
    | (length argDescrs > length args) = Left ("Not enough arguments.")
    | (length argDescrs < length args) = Left ("Too many arguments.")
    | otherwise = 
            collectErrors $
            map checkArgument (zip argDescrs args)
    where checkArgument :: (ArgDescr a, String) -> Either (ArgDescr a, String) a
          checkArgument (d@(Argument i f), s) = case (f s) of
                                                   (Left e) -> Left (d, e)
                                                   (Right r) -> Right r

          collectErrors :: [Either (ArgDescr a, String) a] -> Either String [a]
          collectErrors l = if (any leftF l) then Left ("Errors found in " ++ show (map getInfo $ filter leftF l))
                                             else rights l
            where leftF :: Either (ArgDescr a, String) a -> Bool
                  leftF (Left _) = True
                  leftF _ = False

                  rights :: [Either (ArgDescr a, String) a] -> Either String [a]
                  rights l = Right (rr l)
                    where rr :: [Either (ArgDescr a, String) a] -> [a]
                          rr [] = []
                          rr ((Left _):ss) = rr ss
                          rr ((Right x):ss) = x:(rr ss)

                  getInfo :: Either (ArgDescr a, String) a -> String
                  getInfo (Left (Argument s _, e)) = "Error in " ++ s ++ ": " ++ e
                  getInfo (Right _) = ""

-- Argument parsing functions
stringArgument :: String -> Either String String
stringArgument s = Right s

intArgument :: String -> Either String Int
intArgument s = readEither s

liftArgument :: (b -> a) 
    -> (String -> Either String b) 
    -> (String -> Either String a)
liftArgument l f = \s -> case (f s) of
                            Left e -> Left e
                            Right x -> Right (l x)

-- predefined commands
showHelp :: Commands a r -> [a] -> IO ()
showHelp (Commands cmds) _ = do outputLn "Help."
                                mapM_ showHelpCmd cmds

getArgInfo :: ArgDescr a -> String
getArgInfo (Argument i _) = i
getArgInfo (RestArgument i _) = i

showHelpCmd :: (String, CommandDescr a r) -> IO ()
showHelpCmd (n, CommandDescr i args _) = do let commandStr = generateCommandStr n args
                                            outputLn (commandStr ++ " - " ++ i)

generateCommandStr :: String -> [ArgDescr a] -> String
generateCommandStr n args = showPredList (n ++ " ") (map getArgInfo args) " " "" ""

unknown :: String -> String -> IO ()
unknown cmd err = do outputLn ("There was a problem with \"" ++ cmd ++ "\":")
                     outputLn (err)
                     outputLn ("Try help.")


