module Test.TestSolve (
    solveCommand
) where

import Data.Maybe
import Data.List
import Test.TestGeneral
import System.Console.GetOpt
import System.IO

import Gersdorf.Data.List
import Gersdorf.ContextMatching.Base.Terms
import Gersdorf.ContextMatching.Base.SimpleTermEquation
import Gersdorf.ContextMatching.General.EnvEquationSet
import Gersdorf.Data.StandardTreePath
import Gersdorf.ContextMatching.Linear.LinearityTest
import Gersdorf.ContextMatching.Linear.LinearCM
import Gersdorf.ContextMatching.Test.TestTermGeneratorParameters

import Test.Command

solveCommand :: Command (Flag, String)
solveCommand = Command { cmdName = "solve", 
                         cmdOptions = solveOptions, 
                         cmdFunction = computeSolution } 
                                         
data Flag =
    FlagProblem
    | OutputPath
    | FlagStrategy
    | FlagHelp
    | FlagType
    | FlagTGP
    deriving (Show, Eq)

solveOptions :: [OptDescr (Flag, String)]
solveOptions =
    [ Option ['p'] ["problem"]  (ReqArg (att FlagProblem) "String") 
        ("A String as a unification problem. E. g. \"[ C[f(x)] =? g(f(a()), f(b())) ]\".\n"
        ++ "The problem \"random\" is treated as a randomly generated Problem. See -r.")
    , Option ['o'] ["output"]      (ReqArg (att OutputPath) "FILE")
        "output path. Default: stdout"
    , Option ['t'] ["type"]    (ReqArg (att FlagType) "normal|general|linear|verbose|statistic|islinear")
        "Type of calculation. Default: normal"
    , Option ['s'] ["strategy"]    (ReqArg (att FlagStrategy) "extendedStrategy|extended2Strategy|paperStrategy|noSplitStrategy|baseStrategy")
        "Strategy. Default: baseStrategy."
    , Option ['r'] ["randomterm"]         (ReqArg (att FlagTGP) (concat $ intersperse "|" $ map fst termGeneratorParameters)) 
        "Parameter to Termgenerator. Default: smallRTP."
    , Option ['h'] ["help"]        (NoArg (FlagHelp, ""))
        "Shows help"
    ]
    where
    att :: Flag -> String -> (Flag, String)
    att f v = (f, v)

getOption :: [(Flag, String)] -> (Flag -> String) -> Flag -> String
getOption flags defaultProvider f =
    defaultize (lookup f flags)
    where
    defaultize Nothing = defaultProvider f
    defaultize (Just s) = s

getOptionInt :: [(Flag, String)] -> (Flag -> String) -> Flag -> Int
getOptionInt flags defaultProvider f =
    read $ getOption flags defaultProvider f

getOptionBool :: [(Flag, String)] -> (Flag -> String) -> Flag -> Bool
getOptionBool flags defaultProvider f =
    read $ getOption flags defaultProvider f

defaultOptions :: Flag -> String
defaultOptions f 
    | (f == OutputPath)  = "-"
    | (f == FlagType) = "normal"
    | (f == FlagTGP)  = "smallRTP"
    | otherwise          = ""

----------------------------------------------------------------------------
--------------------------------------------------------------------

getStrategy :: (EnvPosEqSet SimpleEnvEquationSet eq t tp) =>
    [(Flag, String)] 
    -> [Strategy SimpleEnvEquationSet eq t tp]
getStrategy flags =
    let strategyNames = map snd $ filter ((==) FlagStrategy . fst) flags
        findStrategy name =
            let result = filter ((==) name . strategyName) strategies
            in  if (null result)
                    then Nothing
                    else Just (head result)
    in  mapRemove findStrategy strategyNames

termGeneratorParameters :: [(String, TermGeneratorParameter)]
termGeneratorParameters =
    [("varRTP", varRTP), ("defaultRTP", defaultRTP),
     ("holeRTP", holeRTP), ("smallRTP", smallRTP),
     ("arityRTP", arityRTP), ("cvarRTP", cvarRTP)]

getTermGeneratorParameter :: 
    String 
    -> TermGeneratorParameter
getTermGeneratorParameter name = 
    case (lookup name termGeneratorParameters) of
        Nothing -> smallRTP
        (Just rtp) -> rtp

computeSolution :: [(Flag, String)] -> IO ()
computeSolution flags =
    let outputPath  = getOption flags defaultOptions OutputPath
        calctype    = getOption flags defaultOptions FlagType
        pStr = getOption flags defaultOptions FlagProblem
        tgp  = getTermGeneratorParameter $ getOption flags defaultOptions FlagTGP

        strategy= let configuredStrs = getStrategy flags
                  in  if (null configuredStrs)
                        then baseStrategy
                        else head $ configuredStrs

        help = case (lookup FlagHelp flags) of
                    Nothing -> False
                    (Just _) -> True

    in  do if (help) 
             then putStr $ usageInfo "Usage of command solve" solveOptions
             else do putStr "Usage help with --help.\n"
                     problem     <-
                       do if (pStr == "random")
                              then generateUP tgp
                              else case (parseUPError pStr) of
                                        (Left error) -> ioError (userError ("Could not parse Problem: " ++ show error))
                                        (Right up)   -> return up
                              
                     let strPars = ("Computed solutions with:"
                                    ++ "\n  outputPath    = " ++ outputPath
                                    ++ "\n  strategy      = " ++ show strategy
                                    ++ "\n  problem       = " ++ show problem
                                    ++ "\n  type          = " ++ calctype
                                    ++ "\n")

                     putStr strPars
                             
                     let solutionStrings = solve calctype strategy problem

                     if (outputPath == "-")
                        then mapM_ (\s -> putStrLn s) solutionStrings
                        else do h <- openFile outputPath WriteMode
                                hPutStr h strPars
                                mapM_ (\s -> do { putStrLn s; hPutStrLn h s }) solutionStrings


solve :: 
    String
    -> Strategy SimpleEnvEquationSet SimpleTermEquation ContextTerm StandardTreePath
    -> ContextUP
    -> [String]
solve calctype strategy up
    | calctype == "general" = map show $ solveGCM strategy up
    | calctype == "verbose" = map show $ solveGCMVerbose strategy up
    | calctype == "statistic" = [show $ solveGCMStatistic strategy up]
    | calctype == "islinear"  = [show $ isLinear up]
    | calctype == "linear"    = 
        if (isLinear up)
            then map show $ solveLinearCM up
            else ["Problem " ++ show up ++ " is not linear."]
    | calctype == "normal"    =
        if (isLinear up)
            then ("Using linear algorithm.\n"):(map show $ solveLinearCM up)
            else ("Using general algorithm.\n"):(map show $ solveGCM strategy up)
    | otherwise = ["Unknown type."]
