module Test.TestStatistic (
    statisticCommand
) where

import Data.Maybe
import Data.List
import Test.TestGeneral
import System.Console.GetOpt
import System.IO

import Gersdorf.Data.List
import Gersdorf.ContextMatching.Test.TestTermGeneratorParameters
import Gersdorf.ContextMatching.Test.StatisticTests

import Test.Command

statisticCommand :: Command (Flag, String)
statisticCommand = Command { cmdName = "statistic", 
                             cmdOptions = statisticOptions, 
                             cmdFunction = computeStatistic } 
                                         
data Flag =
    SampleSize 
    | SampleCount 
    | FlagTGP
    | OutputPath
    | FlagStrategy
    | FlagHelp
    deriving (Show, Eq)

statisticOptions :: [OptDescr (Flag, String)]
statisticOptions =
    [ Option ['n'] ["samplesize"]  (ReqArg (att SampleSize) "INT") 
        "sample size. Default 50"
    , Option ['c'] ["samplecount"] (ReqArg (att SampleCount) "INT") 
        "sample count. Default 20"
    , Option ['r'] ["randomterm"]         (ReqArg (att FlagTGP) (concat $ intersperse "|" $ map fst termGeneratorParameters)) 
        "Parameter to Termgenerator. Default: smallRTP."
    , Option ['o'] ["output"]      (ReqArg (att OutputPath) "FILE")
        "output path. Default: stdout"
    , Option ['s'] ["strategy"]    (ReqArg (att FlagStrategy) "extendedStrategy|extended2Strategy|paperStrategy|noSplitStrategy|baseStrategy")
        "Strategy. Default: baseStrategy."
    , Option ['h'] ["help"]        (NoArg (FlagHelp, ""))
        "Shows help"
    ]
    where
    att :: Flag -> String -> (Flag, String)
    att f v = (f, v)

getOption :: [(Flag, String)] -> (Flag -> String) -> Flag -> String
getOption flags defaultProvider f =
    defaultize (lookup f flags)
    where
    defaultize Nothing = defaultProvider f
    defaultize (Just s) = s

getOptionInt :: [(Flag, String)] -> (Flag -> String) -> Flag -> Int
getOptionInt flags defaultProvider f =
    read $ getOption flags defaultProvider f

defaultOptions :: Flag -> String
defaultOptions f 
    | (f == SampleSize)  = "50"
    | (f == SampleCount) = "20"
    | (f == OutputPath)  = "-"
    | (f == FlagTGP)     = "smallRTP"
    | otherwise          = ""

----------------------------------------------------------------------------
--------------------------------------------------------------------

termGeneratorParameters :: [(String, TermGeneratorParameter)]
termGeneratorParameters =
    [("varRTP", varRTP), ("defaultRTP", defaultRTP),
     ("holeRTP", holeRTP), ("smallRTP", smallRTP),
     ("arityRTP", arityRTP), ("cvarRTP", cvarRTP)]

getTermGeneratorParameter :: 
    String 
    -> TermGeneratorParameter
getTermGeneratorParameter name = 
    case (lookup name termGeneratorParameters) of
        Nothing -> smallRTP
        (Just rtp) -> rtp

getStrategies :: (EnvPosEqSet SimpleEnvEquationSet eq t tp) =>
    [(Flag, String)] 
    -> [Strategy SimpleEnvEquationSet eq t tp]
getStrategies flags =
    let strategyNames = map snd $ filter ((==) FlagStrategy . fst) flags
        findStrategy name =
            let result = filter ((==) name . strategyName) strategies
            in  if (null result)
                    then Nothing
                    else Just (head result)
    in  mapRemove findStrategy strategyNames

computeStatistic :: [(Flag, String)] -> IO ()
computeStatistic flags =
    let sampleSize  = getOptionInt flags defaultOptions SampleSize
        sampleCount = getOptionInt flags defaultOptions SampleCount
        outputPath  = getOption flags defaultOptions OutputPath
        tgp  = getTermGeneratorParameter $ getOption flags defaultOptions FlagTGP
        strs = let configuredStrs = getStrategies flags
               in  if (null configuredStrs)
                       then strategies
                       else configuredStrs

        help = case (lookup FlagHelp flags) of
                    Nothing -> False
                    (Just _) -> True

        outputter maybeH lineIO =
           do line <- lineIO
              case (maybeH) of
                Nothing -> putStr line
                (Just h) -> do putStr line
                               hPutStr h line
                               hFlush h
 
    in  do if (help) 
             then putStr $ usageInfo "Usage of command statistic" statisticOptions
             else do putStr "Usage help with --help.\n"
                     let strPars = ("Computed statistics with:"
                                    ++ "\n  sampleSize    = " ++ show sampleSize
                                    ++ "\n  sampleCount   = " ++ show sampleCount
                                    ++ "\n  outputPath    = " ++ show outputPath
                                    ++ "\n  strategies    = " ++ show strs
                                    ++ "\n  termGenerator = " ++ show tgp
                                    ++ "\n")
                         upsG = generateUPn tgp sampleSize

                     putStr strPars
                     let statisticsStrsIO = createStatistics strs upsG sampleCount

                     if (outputPath == "-")
                        then do let outputActions = map (outputter Nothing) statisticsStrsIO
                                sequence_ outputActions
                        else do h <- openFile outputPath AppendMode
                                hPutStr h strPars
                                let outputActions = map (outputter (Just h)) statisticsStrsIO
                                sequence_ outputActions
      

